# -*- coding: utf-8 -*-
# File: setup.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import re
import sys

from setuptools import find_packages, setup

ROOT = os.path.realpath(os.path.join(os.path.dirname(__file__)))

with open(os.path.join(ROOT, "README.md"), "rb") as f:
    long_description = f.read().decode("utf-8")


def get_version():
    init_path = os.path.join(ROOT, "deepdoctection", "__init__.py")
    init_py = open(init_path, "r").readlines()
    version_line = [l.strip() for l in init_py if l.startswith("__version__")][0]
    version = version_line.split("=")[-1].strip().strip("'\"")
    return version


sys.path.insert(0, ROOT)

# Taken from https://github.com/huggingface/transformers/blob/master/setup.py. Will list all dependencies, even those
# that need to be installed separately
_DEPS = [
    # the minimum requirements to run pipelines without considering DL models specific dependencies
    "apted==1.0.3",
    "catalogue==2.0.7",
    "distance==0.1.3",
    "huggingface_hub>=0.12.0",
    "importlib-metadata>=4.11.2",
    "jsonlines==3.0.0",
    "lxml>=4.9.1",
    "mock==4.0.3",
    "networkx>=2.7.1",
    "numpy>=1.21,<1.24",
    "opencv-python==4.5.4.60",
    "packaging>=20.0,<22.0",
    "pycocotools>=2.0.2",
    "pypdf2>=1.27.5,<2.10.1",
    "python-prctl",
    "pyyaml==6.0",
    "pyzmq>=16",
    "termcolor>=1.1",
    "tabulate>=0.7.7",
    "tqdm==4.64.0",
    # type-stubs
    "types-PyYAML",
    "types-termcolor==1.1.3",
    "types-tabulate",
    "types-tqdm",
    "lxml-stubs",
    # Tensorflow related dependencies
    "protobuf==3.20.1",
    "tensorpack",
    # PyTorch related dependencies
    "timm",
    "transformers<=4.24",
    "detectron2 @ git+https://github.com/facebookresearch/detectron2.git",
    # other third party related dependencies (services or DL libraries). Must be installed by users
    "jdeskew",
    "boto3",
    "pdfplumber>=0.7.1",
    "tensorflow-addons>=0.13.0",
    "python-doctr<=0.5.1",
    "fasttext",
    # dev dependencies
    "click==8.0.4",  # version will not break black
    "black==22.3.0",
    "isort",
    "pylint==2.13.4",
    "mypy==0.942",
    # docs
    "jinja2==3.0.3",
    "mkdocs-material",
    "mkdocstrings-python",
    "griffe==0.25.0",
    # test
    "pytest",
    "pytest-cov",
    "wandb",
]

# lookup table with items like:
# pycocotools: "pycocotools>=2.0.2"
# tensorpack: "tensorpack"
deps = {b: a for a, b in (re.findall(r"^(([^!=<>]+)(?:[!=<>].*)?$)", x)[0] for x in _DEPS)}


def deps_list(*pkgs: str):
    return [deps[pkg] for pkg in pkgs]


# pypi dependencies without considering DL models specific dependencies
dist_deps = deps_list(
    "apted",
    "catalogue",
    "distance",
    "huggingface_hub",
    "importlib-metadata",
    "jsonlines",
    "lxml",
    "mock",
    "networkx",
    "numpy",
    "opencv-python",
    "packaging",
    "pycocotools",
    "pypdf2",
    "pyyaml",
    "pyzmq",
    "termcolor",
    "tabulate",
    "tqdm",
    "types-PyYAML",
    "types-termcolor",
    "types-tabulate",
    "types-tqdm",
    "lxml-stubs",
)


# remaining dependencies to use models that neither require TF nor PyTorch
additional_deps = deps_list("boto3", "pdfplumber", "fasttext", "jdeskew")

# Tensorflow dependencies
tf_deps = deps_list("tensorpack", "protobuf", "tensorflow-addons", "python-doctr")

# PyTorch dependencies
pt_deps = deps_list("timm", "transformers", "python-doctr")
source_pt_deps = pt_deps + deps_list("detectron2 @ git+https://github.com/facebookresearch/detectron2.git")

# Putting all together
tf_deps = dist_deps + tf_deps + additional_deps
pt_deps = dist_deps + pt_deps + additional_deps
source_pt_deps = dist_deps + source_pt_deps + additional_deps


# if sys.platform == "linux":
#    source_pt_deps.extend(deps_list("python-prctl"))

# dependencies for rtd. Only needed to create requirements.txt
docs_deps = deps_list(
    "tensorpack",
    "boto3",
    "transformers",
    "pdfplumber",
    "lxml",
    "lxml-stubs",
    "pycocotools",
    "jdeskew",
    "jinja2",
    "mkdocs-material",
    "mkdocstrings-python",
    "griffe",
)


# test dependencies
test_deps = deps_list("pytest", "pytest-cov")

# dev dependencies
dev_deps = deps_list("click", "black", "isort", "pylint", "mypy", "wandb")

# TODO: add function that lists correct not pre-installed third party libs in package, such that requirement errors
#  can be printed with correct version dependencies.
# when uploading to pypi first comment all source extra dependencies so that there are no dependencies to dataflow

EXTRA_DEPS = {
    "tf": tf_deps,
    "pt": pt_deps,
#    "source-pt": source_pt_deps,
    "docs": docs_deps,
    "dev": dev_deps,
    "test": test_deps,
    "hf": pt_deps,
}

setup(
    name="deepdoctection",
    version=get_version(),
    author="Dr. Janis Meyer",
    url="https://github.com/deepdoctection/deepdoctection",
    license="Apache License 2.0",
    description="Repository for Document AI",
    long_description=long_description,
    long_description_content_type="text/markdown",
    install_requires=dist_deps,
    extras_require=EXTRA_DEPS,
    packages=find_packages(),
    package_data={
        "deepdoctection.configs": ["*.yaml"],
        "deepdoctection.datasets.instances.xsl": ["*.xsl"],
        "deepdoctection": ["py.typed"],
    },
    include_package_data=True,
    classifiers=[
        "Development Status :: 4 - Beta",
        "License :: OSI Approved :: Apache Software License",
        "Natural Language :: English",
        "Operating System :: POSIX :: Linux",
        "Programming Language :: Python :: 3.8",
        "Topic :: Scientific/Engineering :: Artificial Intelligence",
    ],
    python_requires=">=3.8",
)



# -*- coding: utf-8 -*-
# File: test_tokenclass.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module pipe.tokenclass
"""

from typing import List
from unittest.mock import MagicMock

from pytest import mark

from deepdoctection.datapoint import Image
from deepdoctection.extern.base import SequenceClassResult, TokenClassResult
from deepdoctection.pipe import LMSequenceClassifierService, LMTokenClassifierService
from deepdoctection.utils.file_utils import transformers_available
from deepdoctection.utils.settings import BioTag, PageType, TokenClasses, WordType

if transformers_available():
    from transformers import LayoutLMForSequenceClassification, LayoutLMForTokenClassification, LayoutLMTokenizerFast


class TestLMTokenClassifierService:
    """
    Test LMTokenClassifierService
    """

    @staticmethod
    @mark.requires_pt
    def test_pass_datapoint_2(
        dp_image_with_layout_and_word_annotations: Image,
        token_class_result: List[TokenClassResult],
    ) -> None:
        """
        Testing pass_datapoint with fast tokenizer
        """

        # Arrange
        tokenizer_fast = LayoutLMTokenizerFast.from_pretrained("microsoft/layoutlm-base-uncased")

        language_model = MagicMock()
        language_model.predict = MagicMock(return_value=token_class_result)
        language_model.name = "test"
        language_model.default_kwargs_for_input_mapping = MagicMock(return_value={})
        language_model.model = MagicMock(spec=LayoutLMForTokenClassification)
        language_model.model.config = MagicMock()
        language_model.model.config.tokenizer_class = None
        lm_service = LMTokenClassifierService(tokenizer_fast, language_model)

        dp = dp_image_with_layout_and_word_annotations

        # Act
        dp = lm_service.pass_datapoint(dp)

        # Assert
        words = dp.get_annotation(annotation_ids="7e79459d-0cf9-3954-802c-6d8c4d017792")
        assert words[0].get_sub_category(WordType.token_class).category_name == TokenClasses.header
        assert words[0].get_sub_category(WordType.tag).category_name == BioTag.begin

        words = dp.get_annotation(annotation_ids="08e645c4-f28d-30e6-befe-ab22dc53da92")
        assert words[0].get_sub_category(WordType.token_class).category_name == TokenClasses.header
        assert words[0].get_sub_category(WordType.tag).category_name == BioTag.begin

        words = dp.get_annotation(annotation_ids="224d479a-77e1-3c7a-9cd1-08430c61f020")
        assert words[0].get_sub_category(WordType.token_class).category_name == TokenClasses.header
        assert words[0].get_sub_category(WordType.tag).category_name == BioTag.inside

        words = dp.get_annotation(annotation_ids="841c74d7-45b2-3e0d-a8d5-7be3391685d6")
        assert words[0].get_sub_category(WordType.token_class).category_name == TokenClasses.header
        assert words[0].get_sub_category(WordType.tag).category_name == BioTag.inside


class TestLMSequenceClassifierService:
    """
    Test LMSequenceClassifierService
    """

    @staticmethod
    @mark.requires_pt
    def test_pass_datapoint(
        dp_image_with_layout_and_word_annotations: Image,
        sequence_class_result: SequenceClassResult,
    ) -> None:
        """
        Testing pass_datapoint
        """

        # Arrange
        tokenizer_fast = LayoutLMTokenizerFast.from_pretrained("microsoft/layoutlm-base-uncased")

        language_model = MagicMock()
        language_model.model = MagicMock(spec=LayoutLMForSequenceClassification)
        language_model.model.config = MagicMock()
        language_model.model.config.tokenizer_class = None
        language_model.predict = MagicMock(return_value=sequence_class_result)
        lm_service = LMSequenceClassifierService(tokenizer_fast, language_model)

        dp = dp_image_with_layout_and_word_annotations

        # Act
        dp = lm_service.pass_datapoint(dp)
        assert dp.summary is not None

        # Assert
        assert dp.summary.get_sub_category(PageType.document_type).category_name == "FOO"



# -*- coding: utf-8 -*-
# File: test_annotation.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Testing the module datapoint.annotation
"""
import pytest
from pytest import mark

from deepdoctection.datapoint import CategoryAnnotation
from deepdoctection.utils import get_uuid
from deepdoctection.utils.settings import get_type

from .conftest import CatAnn


class TestCategoryAnnotation:
    """
    Testing Category Annotation methods
    """

    @staticmethod
    @mark.basic
    @mark.parametrize(
        "external_id,expected",
        [
            (CatAnn.external_id, CatAnn.get_annotation_id("n")),
            (CatAnn.uuid, CatAnn.get_annotation_id("u")),
        ],
    )
    def test_annotation_id_is_correctly_assigned(external_id: str, expected: str) -> None:
        """
        Annotation id is assigned as expected, provided that it is passed as external id
        :param external_id: external_id
        :param expected: annotation_id
        """

        # Arrange
        test_cat = CategoryAnnotation(category_name="FOO", category_id="1", external_id=external_id)

        # Assert
        assert test_cat.annotation_id == expected

    @staticmethod
    @mark.basic
    def test_dump_sup_cat_and_check_ann_id() -> None:
        """
        Sub categories are dumped to category instance and annotation ids are correctly assigned
        """

        # Arrange
        cat = CategoryAnnotation(category_name="FOO", category_id="1")
        sub_cat_1 = CategoryAnnotation(category_name="BAK", category_id="2")
        sub_cat_2 = CategoryAnnotation(category_name="BAZ", category_id="3")

        # Act
        cat.dump_sub_category(get_type("bak"), sub_cat_1)
        cat.dump_sub_category(get_type("baz"), sub_cat_2, "c822f8c3-1148-30c4-90eb-cb4896b1ebe5")

        export_sub_cat_1 = cat.get_sub_category(get_type("bak"))
        export_sub_cat_2 = cat.get_sub_category(get_type("baz"))

        # Assert
        if export_sub_cat_1 is not None and export_sub_cat_2 is not None:
            tmp_annotation_id_cat = "8295618a-978d-3aad-bea9-b4faa3061ab6"
            assert export_sub_cat_1.annotation_id == get_uuid("TestType.BAK2" + tmp_annotation_id_cat)
            assert export_sub_cat_2.annotation_id == get_uuid(
                "TestType.BAZ3" + tmp_annotation_id_cat, "c822f8c3-1148-30c4-90eb-cb4896b1ebe5"
            )

    @staticmethod
    @mark.basic
    def test_dump_sub_cat_with_external_id() -> None:
        """
        Sub categories are dumped to category instance and external category ids will be annotation ids
        """

        # Arrange
        cat = CategoryAnnotation(category_name="FOO", category_id="1")
        sub_cat_1 = CategoryAnnotation(
            category_name="FOOBAK", category_id="2", external_id="c822f8c3-1148-30c4-90eb-cb4896b1ebe5"
        )
        # Act
        cat.dump_sub_category(get_type("bak"), sub_cat_1)

        export_sub_cat_1 = cat.get_sub_category(get_type("bak"))

        # Assert
        if export_sub_cat_1 is not None:
            assert export_sub_cat_1.annotation_id == "c822f8c3-1148-30c4-90eb-cb4896b1ebe5"

    @staticmethod
    @mark.basic
    def test_remove_sub_cat() -> None:
        """
        Existing sub categories are correctly removed
        """

        # Arrange
        cat = CategoryAnnotation(category_name="FOO", category_id="1")
        sub_cat_1 = CategoryAnnotation(category_name="BAK", category_id="2")
        cat.dump_sub_category(get_type("bak"), sub_cat_1)

        # Act
        cat.remove_sub_category(get_type("bak"))

        # Assert
        with pytest.raises(Exception):
            cat.get_sub_category(get_type("bak"))

    @staticmethod
    @mark.basic
    def test_state_id() -> None:
        """
        state_id is correctly determined based by given state attributes
        """

        # Arrange
        cat = CategoryAnnotation(
            category_name="FOO", category_id="1", external_id="c822f8c3-1148-30c4-90eb-cb4896b1e222"
        )
        sub_cat_1 = CategoryAnnotation(
            category_name="FOOBAK", category_id="2", external_id="c822f8c3-1148-30c4-90eb-cb4896b1ebe5"
        )
        # Act
        cat.dump_sub_category(get_type("bak"), sub_cat_1)

        # Assert
        assert cat.state_id == "8301eef5-bd4a-3c10-9dad-7cae94dd150b"



# -*- coding: utf-8 -*-
# File: prodigystruct.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for mapping annotations to and from prodigy data structure
"""

import os
from typing import Mapping, Optional

from ..datapoint import BoundingBox, Image, ImageAnnotation
from ..utils.detection_types import JsonDict, Pathlike
from .maputils import MappingContextManager, curry, maybe_get_fake_score

_PRODIGY_IMAGE_PREFIX = "data:image/png;base64,"


@curry
def prodigy_to_image(
    dp: JsonDict,
    categories_name_as_key: Mapping[str, str],
    load_image: bool,
    fake_score: bool,
    path_reference_ds: Optional[Pathlike] = None,
    accept_only_answer: bool = False,
    category_name_mapping: Optional[Mapping[str, str]] = None,
) -> Optional[Image]:
    """
    Map a datapoint of annotation structure as given as from Prodigy database to an Image
    structure.

    :param dp: A datapoint in dict structure as returned from Prodigy database
    :param categories_name_as_key: A dict of categories, e.g. DatasetCategories.get_categories(name_as_key=True)
    :param load_image: If 'True' it will load image to attr:`Image.image`
    :param fake_score: If dp does not contain a score, a fake score with uniform random variables in (0,1)
                       will be added.
    :param path_reference_ds: A path to a reference-dataset. It must point to the basedir where the file
                              of the datapoint can be found.
    :param accept_only_answer: Filter every datapoint that has the answer 'reject' or 'ignore'.
    :param category_name_mapping: Map incoming category names, e.g. {"source_name":"target_name"}
    :return: Image
    """

    if accept_only_answer and dp.get("answer") != "accept":
        return None

    file_name: Optional[str] = None
    meta = dp.get("meta")
    if meta:
        file_name = meta.get("file")

    if not file_name:
        file_name = dp.get("id")
    if not file_name:
        file_name = dp.get("text")
    if not file_name:
        path = dp.get("path")
        if path:
            path, file_name = os.path.split(path)
    if not file_name:
        file_name = ""

    external_id = dp.get("image_id")

    if not external_id:
        external_id = file_name

    with MappingContextManager(file_name) as mapping_context:
        path_reference_location = ""
        if path_reference_ds:
            path_reference_location = os.path.join(path_reference_ds, file_name)
        location = dp.get("path", path_reference_location)
        if not os.path.isfile(location):
            location = None

        image = Image(file_name=file_name, location=location, external_id=external_id)

        if dp.get("image"):
            image.image = dp["image"].split(",")[1]
            if not load_image:
                image.clear_image()
        elif "width" in dp and "height" in dp:
            image.set_width_height(dp["width"], dp["height"])

        spans = dp.get("spans", [])

        for span in spans:
            ulx, uly = list(map(float, span["points"][0]))
            lrx, lry = list(map(float, span["points"][2]))
            ulx = min(max(ulx, 0), image.width if image.width else ulx)
            uly = min(max(uly, 0), image.height if image.height else uly)
            lrx = min(max(lrx, 0), image.width if image.width else lrx)
            lry = min(max(lry, 0), image.height if image.height else lry)
            upper_left = [ulx, uly]
            lower_right = [lrx, lry]

            bbox = BoundingBox(
                absolute_coords=True, ulx=upper_left[0], uly=upper_left[1], lrx=lower_right[0], lry=lower_right[1]
            )
            external_id = span.get("annotation_id")
            if not external_id:
                external_id = span.get("id")

            score = span.get("score")

            if not score:
                score = maybe_get_fake_score(fake_score)

            if category_name_mapping is not None:
                label = category_name_mapping.get(span["label"])
                if not label:
                    label = span["label"]
            else:
                label = span["label"]
            if not isinstance(label, str):
                raise ValueError("label could not assigned to be a string")

            annotation = ImageAnnotation(
                category_name=label,
                bounding_box=bbox,
                category_id=categories_name_as_key[label],
                score=score,
                external_id=external_id,
            )
            image.dump(annotation)

    if mapping_context.context_error:
        return None
    return image


def image_to_prodigy(dp: Image) -> JsonDict:
    """
    The mapper to transform the normalized image representation of datasets into the format
    for visualising the annotation components in Prodigy.

    :param dp: An image
    :return: A dictionary with compulsory keys: "text" and "spans"
    """

    assert isinstance(dp, Image), f"datapoints must be of type Image, is of type {type(dp)}"
    output: JsonDict = {}
    img_str = dp.get_image().to_b64()
    if img_str is not None:
        output["image"] = _PRODIGY_IMAGE_PREFIX + img_str
    output["text"] = dp.file_name
    output["image_id"] = dp.image_id

    spans = []
    for ann in dp.get_annotation_iter():
        if ann.bounding_box is None:
            raise ValueError("ann.bounding_box cannot be None")
        box: JsonDict = {"label": ann.category_name, "annotation_id": ann.annotation_id}
        if ann.score is not None:
            box["score"] = float(round(ann.score, 3))
        box["type"] = "rect"
        if ann.image is not None:
            bounding_box = ann.image.get_embedding(dp.image_id)
        else:
            bounding_box = ann.bounding_box
        if not bounding_box.absolute_coords:
            bounding_box = bounding_box.transform(dp.width, dp.height, absolute_coords=True)
        boxes = [
            [bounding_box.ulx, bounding_box.uly],
            [bounding_box.ulx, bounding_box.lry],
            [bounding_box.lrx, bounding_box.lry],
            [bounding_box.lrx, bounding_box.uly],
        ]

        box["points"] = [list(map(float, box)) for box in boxes]

        spans.append(box)
    output["spans"] = spans
    output["meta"] = {"file_name": dp.file_name}
    return output



# -*- coding: utf-8 -*-
# File: test_funsd.py

# Copyright 2022 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module datasets.instances.funsd
"""

from unittest.mock import MagicMock, patch

from pytest import mark

from deepdoctection.datasets import Funsd
from deepdoctection.utils.settings import LayoutType, WordType

from ...test_utils import collect_datapoint_from_dataflow, get_test_path
from .conftest import get_white_image


@mark.basic
@patch("deepdoctection.mapper.xfundstruct.load_image_from_file", MagicMock(side_effect=get_white_image))
def test_dataset_funsd_returns_image_and_annotations() -> None:
    """
    test dataset funsd returns image and annotations
    """

    # Arrange
    funsd = Funsd()
    funsd.dataflow.get_workdir = get_test_path  # type: ignore
    funsd.dataflow.splits = {"test": ""}
    funsd.dataflow.annotation_files = {"test": ""}
    df = funsd.dataflow.build()

    # Act
    df_list = collect_datapoint_from_dataflow(df)
    assert len(df_list) == 3  # the first three images coming from files not related to funsd data
    dp = df_list[2]
    word = dp.get_annotation(category_names=LayoutType.word)[0]
    assert word.get_sub_category(WordType.token_class) is not None
    assert word.get_sub_category(WordType.characters) is not None
    assert word.get_sub_category(WordType.tag) is not None
    assert word.get_sub_category(WordType.token_tag) is not None

    text = dp.get_annotation(category_names=LayoutType.text)[0]
    assert text.get_sub_category(WordType.token_class) is not None



# -*- coding: utf-8 -*-
# File: test_language.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module pipe.language
"""

from unittest.mock import MagicMock

from pytest import mark

from deepdoctection.datapoint import ContainerAnnotation, Image
from deepdoctection.extern.base import DetectionResult
from deepdoctection.pipe.language import LanguageDetectionService
from deepdoctection.pipe.text import TextOrderService
from deepdoctection.utils import CellType, LayoutType, PageType


class TestLanguageDetectionService:
    """
    Test LanguageDetectionService
    """

    def setup_method(self) -> None:
        """
        setup necessary components
        """

        self._language_detector = MagicMock()
        self._text_order_service = TextOrderService(
            text_container=LayoutType.word,
            floating_text_block_names=[LayoutType.title, LayoutType.text, LayoutType.list],
            text_block_names=[
                LayoutType.title,
                LayoutType.text,
                LayoutType.list,
                LayoutType.cell,
                CellType.header,
                CellType.body,
            ],
        )
        self.language_detection_service = LanguageDetectionService(
            self._language_detector,
            text_container=LayoutType.word,
            text_block_names=[
                LayoutType.title,
                LayoutType.text,
                LayoutType.list,
                LayoutType.cell,
                CellType.header,
                CellType.body,
            ],
        )

    @mark.basic
    def test_pass_datapoint(
        self, dp_image_with_layout_and_word_annotations: Image, language_detect_result: DetectionResult
    ) -> None:
        """
        test pass datapoint
        """
        # Arrange
        dp_image = dp_image_with_layout_and_word_annotations
        self._language_detector.predict = MagicMock(return_value=language_detect_result)
        dp_with_text_ordered = self._text_order_service.pass_datapoint(dp_image)  # need to setup a reading order

        # Act
        dp = self.language_detection_service.pass_datapoint(dp_with_text_ordered)

        # Assert
        assert dp.summary is not None
        assert dp.summary.get_sub_category(PageType.language).category_name == "language"
        container_ann = dp.summary.get_sub_category(PageType.language)
        assert isinstance(container_ann, ContainerAnnotation)
        assert container_ann.value == "eng"



# -*- coding: utf-8 -*-
# File: conftest.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Fixtures for extern package testing
"""
from typing import List

from pytest import fixture

from deepdoctection.utils.detection_types import JsonDict

from ..data import get_textract_response
from ..mapper.data import DatapointXfund
from .data import PDF_BYTES, PDF_BYTES_2, get_detr_categories


@fixture(name="layoutlm_input_for_predictor")
def fixture_layoutlm_input_for_predictor() -> JsonDict:
    """
    Layoutlm input
    """
    return DatapointXfund().get_layout_input()


@fixture(name="layoutlm_v2_input")
def fixture_layoutlm_input() -> JsonDict:
    """
    Layoutlm_v2 input
    """
    return DatapointXfund().get_layout_v2_input()


@fixture(name="categories_semantics")
def fixture_categories_semantics() -> List[str]:
    """
    Categories semantics
    """
    return DatapointXfund().get_categories_semantics()


@fixture(name="categories_bio")
def fixture_categories_bio() -> List[str]:
    """
    Categories semantics
    """
    return DatapointXfund().get_categories_bio()


@fixture(name="token_class_names")
def fixture_token_class_names() -> List[str]:
    """
    Categories semantics
    """
    return DatapointXfund().get_token_class_names()


@fixture(name="textract_response")
def fixture_textract_response() -> JsonDict:
    """fixture textract_response"""
    return get_textract_response()


@fixture(name="pdf_bytes")
def fixture_pdf_bytes() -> bytes:
    """
    fixture pdf bytes
    """
    return PDF_BYTES


@fixture(name="pdf_bytes_page_2")
def fixture_pdf_bytes_page_2() -> bytes:
    """
    fixture pdf bytes
    """
    return PDF_BYTES_2


@fixture(name="detr_categories")
def fixture_detr_categories() -> JsonDict:
    """
    fixture object types
    """
    return get_detr_categories()



# -*- coding: utf-8 -*-
# File: test_cocometric.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for metrics that require the COCOeval class.
"""

from copy import copy
from typing import Dict, List, Optional, Tuple, Union

import numpy as np

from ..dataflow import DataFlow
from ..datasets.info import DatasetCategories
from ..mapper.cats import re_assign_cat_ids
from ..mapper.cocostruct import image_to_coco
from ..utils.detection_types import JsonDict
from ..utils.file_utils import Requirement, cocotools_available, get_cocotools_requirement
from .base import MetricBase
from .registry import metric_registry

if cocotools_available():
    from pycocotools.coco import COCO
    from pycocotools.cocoeval import COCOeval

__all__ = ["CocoMetric"]


_COCOEVAL_DEFAULTS = [
    {"ap": 1, "iouThr": "", "areaRng": "all", "maxDets": 100},
    {"ap": 1, "iouThr": 0.5, "areaRng": "all", "maxDets": 100},
    {"ap": 1, "iouThr": 0.75, "areaRng": "all", "maxDets": 100},
    {"ap": 1, "iouThr": "", "areaRng": "small", "maxDets": 100},
    {"ap": 1, "iouThr": "", "areaRng": "medium", "maxDets": 100},
    {"ap": 1, "iouThr": "", "areaRng": "large", "maxDets": 100},
    {"ap": 0, "iouThr": "", "areaRng": "all", "maxDets": 1},
    {"ap": 0, "iouThr": "", "areaRng": "all", "maxDets": 10},
    {"ap": 0, "iouThr": "", "areaRng": "all", "maxDets": 100},
    {"ap": 0, "iouThr": "", "areaRng": "small", "maxDets": 100},
    {"ap": 0, "iouThr": "", "areaRng": "medium", "maxDets": 100},
    {"ap": 0, "iouThr": "", "areaRng": "large", "maxDets": 100},
]

_F1_DEFAULTS = [
    {"ap": 1, "iouThr": 0.9, "areaRng": "all", "maxDets": 100},
    {"ap": 0, "iouThr": 0.9, "areaRng": "all", "maxDets": 100},
]

_MAX_DET_INDEX = [2, 2, 2, 2, 2, 2, 0, 1, 2, 2, 2, 2]

"""
The following function is taken from
https://github.com/cocodataset/cocoapi/blob/master/PythonAPI/pycocotools/cocoeval.py
"""


def _summarize(  # type: ignore
    self, ap: int = 1, iouThr: float = 0.9, areaRng: str = "all", maxDets: int = 100
) -> float:
    # pylint: disable=C0103
    p = self.params
    iStr = " {:<18} {} @[ IoU={:<9} | area={:>6s} | maxDets={:>3d} ] = {:0.3f}"
    titleStr = "Average Precision" if ap == 1 else "Average Recall"
    typeStr = "(AP)" if ap == 1 else "(AR)"
    iouStr = (
        "{:0.2f}:{:0.2f}".format(p.iouThrs[0], p.iouThrs[-1])  # pylint: disable=C0209
        if iouThr is None
        else "{:0.2f}".format(iouThr)  # pylint: disable=C0209
    )

    aind = [i for i, aRng in enumerate(p.areaRngLbl) if aRng == areaRng]
    mind = [i for i, mDet in enumerate(p.maxDets) if mDet == maxDets]
    if ap == 1:
        # dimension of precision: [TxRxKxAxM]
        s = self.eval["precision"]
        # IoU
        if iouThr is not None:
            t = np.where(iouThr == p.iouThrs)[0]
            s = s[t]
        s = s[:, :, :, aind, mind]
    else:
        # dimension of recall: [TxKxAxM]
        s = self.eval["recall"]
        if iouThr is not None:
            t = np.where(iouThr == p.iouThrs)[0]
            s = s[t]
        s = s[:, :, aind, mind]
    if len(s[s > -1]) == 0:
        mean_s = -1
    else:
        mean_s = np.mean(s[s > -1])
    print(iStr.format(titleStr, typeStr, iouStr, areaRng, maxDets, mean_s))
    return mean_s


if cocotools_available():
    COCOeval.summarize_f1 = _summarize


@metric_registry.register("coco")
class CocoMetric(MetricBase):
    """
    Metric induced by `pycocotools.cocoeval.COCOeval`.
    """

    name = "mAP and mAR"
    metric = COCOeval if cocotools_available() else None
    mapper = image_to_coco  # type: ignore
    _f1_score = None
    _f1_iou = None
    _params: Dict[str, Union[List[int], List[List[int]]]] = {}

    @classmethod
    def dump(
        cls, dataflow_gt: DataFlow, dataflow_predictions: DataFlow, categories: DatasetCategories
    ) -> Tuple["COCO", "COCO"]:
        cats = [{"id": int(k), "name": v} for k, v in categories.get_categories(as_dict=True, filtered=True).items()]
        imgs_gt, imgs_pr = [], []
        anns_gt, anns_pr = [], []

        dataflow_gt.reset_state()
        dataflow_predictions.reset_state()

        for dp_gt, dp_pred in zip(dataflow_gt, dataflow_predictions):
            img_gt, ann_gt = cls.mapper(dp_gt)  # type: ignore
            dp_pred = re_assign_cat_ids(  # pylint: disable=E1120
                categories.get_categories(as_dict=True, filtered=True, name_as_key=True)
            )(dp_pred)
            img_pr, ann_pr = cls.mapper(dp_pred)  # type: ignore
            imgs_gt.append(img_gt)
            imgs_pr.append(img_pr)
            anns_gt.extend(ann_gt)
            anns_pr.extend(ann_pr)

        dataset_gt = {"images": imgs_gt, "annotations": anns_gt, "categories": cats}
        dataset_pr = {"images": imgs_pr, "annotations": anns_pr, "categories": cats}

        coco_gt = COCO()
        coco_gt.dataset = dataset_gt
        coco_gt.createIndex()
        coco_predictions = COCO()
        coco_predictions.dataset = dataset_pr
        coco_predictions.createIndex()

        return coco_gt, coco_predictions

    @classmethod
    def get_distance(
        cls, dataflow_gt: DataFlow, dataflow_predictions: DataFlow, categories: DatasetCategories
    ) -> List[JsonDict]:
        coco_gt, coco_predictions = cls.dump(dataflow_gt, dataflow_predictions, categories)

        metric = cls.metric(coco_gt, coco_predictions, iouType="bbox")
        if cls._params:
            for key, value in cls._params.items():
                setattr(metric.params, key, value)

        metric.evaluate()
        metric.accumulate()

        if cls._f1_score:
            summary_bbox = [
                metric.summarize_f1(1, cls._f1_iou, maxDets=metric.params.maxDets[2]),
                metric.summarize_f1(0, cls._f1_iou, maxDets=metric.params.maxDets[2]),
            ]
        else:
            metric.summarize()
            summary_bbox = metric.stats

        results = []
        for params, value in zip(cls.get_summary_default_parameters(), summary_bbox):
            params = copy(params)
            params["mode"] = "bbox"
            params["val"] = value
            results.append(params)

        return results

    @classmethod
    def get_summary_default_parameters(cls) -> List[JsonDict]:
        """
        Returns default parameters of evaluation results. May differ from other CocoMetric classes.

        :return: List of dict with default configuration, e.g. setting of average precision, iou threshold,
                 area range and maximum detections.
        """
        if cls._f1_score:
            if cls._params:
                if cls._params.get("maxDets") is not None:
                    for el, idx in zip(_F1_DEFAULTS, [2, 2]):
                        el["maxDets"] = cls._params["maxDets"][idx]
                        el["iouThr"] = cls._f1_iou
                    return _F1_DEFAULTS
        if cls._params:
            if cls._params.get("maxDets") is not None:
                for el, idx in zip(_COCOEVAL_DEFAULTS, _MAX_DET_INDEX):
                    el["maxDets"] = cls._params["maxDets"][idx]
        return _COCOEVAL_DEFAULTS

    @classmethod
    def set_params(
        cls,
        max_detections: Optional[List[int]] = None,
        area_range: Optional[List[List[int]]] = None,
        f1_score: bool = False,
        f1_iou: float = 0.9,
    ) -> None:
        """
        Setting params for different coco metric modes.

        :param max_detections: The maximum number of detections to consider
        :param area_range: The area range to classify objects as "all", "small", "medium" and "large"
        :param f1_score: Will use f1 score setting with default iouThr 0.9
        :param f1_iou: Use with f1_score True and reset the f1 iou threshold
        """
        if max_detections is not None:
            assert len(max_detections) == 3, max_detections
            cls._params["maxDets"] = max_detections
        if area_range is not None:
            assert len(area_range) == 4, area_range
            cls._params["areaRng"] = area_range

        cls._f1_score = f1_score
        cls._f1_iou = f1_iou

    @classmethod
    def get_requirements(cls) -> List[Requirement]:
        return [get_cocotools_requirement()]



# -*- coding: utf-8 -*-
# File: base.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Module for the base class for building pipelines
"""
from abc import ABC, abstractmethod
from collections import defaultdict
from copy import deepcopy
from typing import Any, Callable, DefaultDict, Dict, List, Mapping, Optional, Set, Union

from ..dataflow import DataFlow, MapData
from ..datapoint.image import Image
from ..extern.base import ImageTransformer, ObjectDetector, PdfMiner, TextRecognizer
from ..mapper.laylmstruct import LayoutLMFeatures
from ..utils.context import timed_operation
from ..utils.detection_types import JsonDict
from .anngen import DatapointManager


class PipelineComponent(ABC):
    """
    Base class for pipeline components. Pipeline components are the parts that make up a pipeline. They contain the
    abstract `serve`, in which the component steps are defined. Within pipelines, pipeline components take an
    image, enrich these with annotations or transform existing annotation and transfer the image again. The pipeline
    component should be implemented in such a way that the pythonic approach of passing arguments via assignment is used
    well. To support the pipeline component, an intrinsic datapoint manager is provided, which can perform operations on
    the image datapoint that are common for pipeline components. This includes the creation of an image, sub-category
    and similar annotations.

    Pipeline components do not necessarily have to contain predictors but can also contain rule-based transformation
    steps. (For pipeline components with predictors see `PredictorPipelineComponent`.)

    The sequential execution of pipeline components is carried out with dataflows. In the case of components with
    predictors, this allows the predictor graph to be set up first and then to be streamed to the processed data points.

    **Caution:** Currently, predictors can only process single images. Processing higher number of batches is not
                 planned.
    """

    def __init__(self, name: str):
        """
        :param name: The name of the pipeline component. The name will be used to identify a pipeline component in a
                     pipeline. Use something that describe the task of the pipeline.
        """
        self.name = name
        self._meta_has_all_types()
        self.dp_manager = DatapointManager()
        self.timer_on = False

    @abstractmethod
    def serve(self, dp: Image) -> None:
        """
        Processing an image through the whole pipeline component. Abstract method that contains all processing steps of
        the component. Please note that dp is already available to the dp_manager and operations for this can be carried
        out via it.

        dp was transferred to the dp_manager via an assignment. This means that operations on dp directly or operations
        via dp_manager are equivalent.

        As a simplified interface `serve` does not have to return a dp. The data point is passed on within
        pipelines internally (via `pass_datapoint`).
        """
        raise NotImplementedError

    def pass_datapoint(self, dp: Image) -> Image:
        """
        Acceptance, handover to dp_manager, transformation and forwarding of dp. To measure the time, use

            self.timer_on = True

        :param dp: datapoint
        :return: datapoint
        """
        if self.timer_on:
            with timed_operation(self.__class__.__name__):
                self.dp_manager.datapoint = dp
                self.serve(dp)
        else:
            self.dp_manager.datapoint = dp
            self.serve(dp)
        return self.dp_manager.datapoint

    def predict_dataflow(self, df: DataFlow) -> DataFlow:
        """
        Mapping a datapoint via `pass_datapoint` within a dataflow pipeline

        :param df: An input dataflow
        :return: A output dataflow
        """
        return MapData(df, self.pass_datapoint)

    @abstractmethod
    def clone(self) -> "PipelineComponent":
        """
        Clone an instance
        """
        raise NotImplementedError

    @abstractmethod
    def get_meta_annotation(self) -> JsonDict:
        """
        Get a dict of list of annotation type. The dict must contain

        `image_annotation` with values: a list of category names,
        `sub_categories` with values: a dict with category names as keys and a list of the generated sub categories
        `relationships` with values: a dict with category names as keys and a list of the generated relationships
        `summaries` with values: A list of summary sub categories
        :return: Dict with meta infos as just described
        """
        raise NotImplementedError

    def _meta_has_all_types(self) -> None:
        if not {"image_annotations", "sub_categories", "relationships", "summaries"}.issubset(
            set(self.get_meta_annotation().keys())
        ):
            raise TypeError(
                f" 'get_meta_annotation' must return dict with all required keys. "
                f"Got {self.get_meta_annotation().keys()}"
            )


class PredictorPipelineComponent(PipelineComponent, ABC):
    """
    Lightweight abstract pipeline component class with `predictor`. Object detectors that only read in images as
    numpy array and return `DetectResult`s are currently permitted.
    """

    def __init__(
        self,
        name: str,
        predictor: Union[ObjectDetector, PdfMiner, TextRecognizer],
    ) -> None:
        """
        :param name: Will be passed to base class
        :param predictor: An Object detector for predicting
        """
        self.predictor = predictor
        super().__init__(name)

    @abstractmethod
    def clone(self) -> "PredictorPipelineComponent":
        raise NotImplementedError


class LanguageModelPipelineComponent(PipelineComponent, ABC):
    """
    Abstract pipeline component class with two attributes `tokenizer` and `language_model` .
    """

    def __init__(
        self,
        name: str,
        tokenizer: Any,
        mapping_to_lm_input_func: Callable[..., Callable[[Image], Optional[LayoutLMFeatures]]],
    ):
        """
        :param name: Will be passed to base class
        :param tokenizer: Tokenizer, typing allows currently anything. This will be changed in the future
        :param mapping_to_lm_input_func: Function mapping image to layout language model features
        """

        self.tokenizer = tokenizer
        self.mapping_to_lm_input_func = mapping_to_lm_input_func
        super().__init__(name)

    @abstractmethod
    def clone(self) -> "LanguageModelPipelineComponent":
        """
        Clone an instance
        """
        raise NotImplementedError


class ImageTransformPipelineComponent(PipelineComponent, ABC):
    """
    Abstract pipeline component class with one model to transform images. This component is meant to be used at the
    beginning of a pipeline
    """

    def __init__(self, name: str, transform_predictor: ImageTransformer):
        """
        :param name: Will be passed to base class
        :param transform_predictor: An `ImageTransformer` for image transformation
        """

        self.transform_predictor = transform_predictor
        super().__init__(name)

    @abstractmethod
    def clone(self) -> "ImageTransformPipelineComponent":
        """
        Clone an instance
        """
        raise NotImplementedError


class Pipeline(ABC):
    """
    Abstract base class for creating pipelines. Pipelines represent the framework with which documents can be processed
    by reading individual pages, processing the pages through the pipeline infrastructure and returning the extracted
    information.

    The infrastructure, as the backbone of the pipeline, consists of a list of pipeline components in which images can
    be passed through via dataflows. The order of the pipeline components in the list determines the processing order.
    The components for the pipeline backbone are composed in `_build_pipe`.

    The pipeline is set up via: `analyze` for a directory with single pages or a document with multiple pages. A
    data flow is returned that is triggered via a for loop and starts the actual processing.

    This creates a pipeline using the following command arrangement:

    **Example:**

            layout = LayoutPipeComponent(layout_detector ...)
            text = TextExtractPipeComponent(text_detector ...)
            simple_pipe = MyPipeline (pipeline_component = [layout, text])
            doc_dataflow = simple_pipe.analyze(input = path / to / dir)

            for page in doc_dataflow:
                print(page)

    In doing so, page contains all document structures determined via the pipeline (either directly from the Image core
    model or already processed further).

    In addition to `analyze`, the internal `_entry` is used to bundle preprocessing steps.
    """

    def __init__(self, pipeline_component_list: List[PipelineComponent]) -> None:
        """
        :param pipeline_component_list: A list of pipeline components.
        """
        self.pipe_component_list = pipeline_component_list

    @abstractmethod
    def _entry(self, **kwargs: Any) -> DataFlow:
        """
        Use this method to bundle all preprocessing, such as loading one or more documents, so that a dataflow is
        provided as a return value that can be passed on to the pipeline backbone.

        :param kwargs: Arguments, for dynamic customizing of the processing or for the transfer of processing types
        """
        raise NotImplementedError

    def _build_pipe(self, df: DataFlow) -> DataFlow:
        """
        Composition of the backbone
        """
        for component in self.pipe_component_list:
            component.timer_on = True
            df = component.predict_dataflow(df)
        return df

    @abstractmethod
    def analyze(self, **kwargs: Any) -> DataFlow:
        """
        Try to keep this method as the only one necessary for the user. All processing steps, such as preprocessing,
        setting up the backbone and post-processing are to be bundled. A dataflow generator df is returned, which is
        generated via

            doc = iter(df)
            page = next(doc)

        can be triggered.
        """
        raise NotImplementedError

    def get_meta_annotation(self) -> JsonDict:
        """
        Collects meta annotations from all pipeline components and summarizes the returned results

        :return: Meta annotations with information about image annotations (list), sub categories (dict with category
                 names and generated sub categories), relationships (dict with category names and generated
                 relationships) as well as summaries (list with sub categories)
        """
        pipeline_populations: Dict[str, Union[List[str], DefaultDict[str, Set[str]]]] = {
            "image_annotations": [],
            "sub_categories": defaultdict(set),
            "relationships": defaultdict(set),
            "summaries": [],
        }
        for component in self.pipe_component_list:
            meta_anns = deepcopy(component.get_meta_annotation())
            pipeline_populations["image_annotations"].extend(meta_anns["image_annotations"])  # type: ignore
            for key, value in meta_anns["sub_categories"].items():
                pipeline_populations["sub_categories"][key].update(value)
            for key, value in meta_anns["relationships"].items():
                pipeline_populations["relationships"][key].update(value)
            pipeline_populations["summaries"].extend(meta_anns["summaries"])  # type: ignore

        return pipeline_populations

    def get_pipeline_info(
        self, position: Optional[int] = None, name: Optional[str] = None
    ) -> Union[Mapping[int, str], str, int]:
        """Get pipeline information: Returns a dictionary with a description of each pipeline component
        :param position: position of the pipeline component in the pipeline
        :param name: name of the pipeline component to search for
        :return: Either a full dictionary with position and name of all pipeline components or the name, if the position
                 has been passed or the position if the name has been passed.
        """
        comp_info = {key: comp.name for key, comp in enumerate(self.pipe_component_list)}
        comp_info_name_as_key = {value: key for key, value in comp_info.items()}
        if position:
            return comp_info[position]
        if name:
            return comp_info_name_as_key[name]
        return comp_info



# -*- coding: utf-8 -*-
# File: tessocr.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Tesseract OCR engine for text extraction
"""
import shlex
import subprocess
import sys
from errno import ENOENT
from itertools import groupby
from os import environ
from typing import Any, Dict, List, Optional, Union

import numpy as np

from ..utils.context import save_tmp_file, timeout_manager
from ..utils.detection_types import ImageType, Requirement
from ..utils.file_utils import _TESS_PATH, TesseractNotFound, get_tesseract_requirement
from ..utils.metacfg import config_to_cli_str, set_config_by_yaml
from ..utils.settings import LayoutType, ObjectTypes
from .base import DetectionResult, ObjectDetector, PredictorBase

# copy and paste with some light modifications from https://github.com/madmaze/pytesseract/tree/master/pytesseract


_LANG_CODE_TO_TESS_LANG_CODE = {
    "fre": "fra",
    "dut": "nld",
    "chi": "chi_sim",
    "cze": "ces",
    "per": "fas",
    "gre": "ell",
    "mac": "mkd",
    "rum": "ron",
    "arm": "hye",
    "geo": "kat",
    "war": "ceb",
    "glg": "gla",
    "slv": "slk",
    "alb": "nor",
    "nn": "eng",
}


class TesseractError(RuntimeError):
    """
    Tesseract Error
    """

    def __init__(self, status: int, message: str) -> None:
        super().__init__()
        self.status = status
        self.message = message
        self.args = (status, message)


def _subprocess_args() -> Dict[str, Any]:
    # See https://github.com/pyinstaller/pyinstaller/wiki/Recipe-subprocess
    # for reference and comments.

    kwargs = {
        "stdin": subprocess.PIPE,
        "stderr": subprocess.PIPE,
        "startupinfo": None,
        "env": environ,
        "stdout": subprocess.PIPE,
    }

    return kwargs


def _input_to_cli_str(lang: str, config: str, nice: int, input_file_name: str, output_file_name_base: str) -> List[str]:
    """
    Generates a tesseract cmd as list of string with given inputs
    """
    cmd_args: List[str] = []

    if not sys.platform.startswith("win32") and nice != 0:
        cmd_args += ("nice", "-n", str(nice))

    cmd_args += (_TESS_PATH, input_file_name, output_file_name_base, "-l", lang)

    if config:
        cmd_args += shlex.split(config)

    cmd_args.append("tsv")

    return cmd_args


def _run_tesseract(tesseract_args: List[str]) -> None:
    try:
        proc = subprocess.Popen(tesseract_args, **_subprocess_args())  # pylint: disable=R1732
    except OSError as error:
        if error.errno != ENOENT:
            raise error from error
        raise TesseractNotFound("Tesseract not found. Please install or add to your PATH.") from error

    with timeout_manager(proc, 0) as error_string:
        if proc.returncode:
            raise TesseractError(
                proc.returncode,
                " ".join(line for line in error_string.decode("utf-8").splitlines()).strip(),  # type: ignore
            )


def image_to_dict(image: ImageType, lang: str, config: str) -> Dict[str, List[Union[str, int, float]]]:
    """
    This is more or less pytesseract.image_to_data with a dict as returned value.
    What happens under the hood is:

    - saving an image file
    - defining tesseracts command line
    - saving a temp .tsv file with predicted results
    - reading the .tsv file and returning the results as dict.

    Requires Tesseract 3.05+

    :param image: Image in np.array.
    :param lang: String of language
    :param config: string of configs
    :return: Dictionary with keys 'left', 'top', 'width', 'height' (bounding box coords), 'conf' (confidence), 'text'
             (captured text), 'block_num' (block number) and 'lin_num' (line number).
    """

    with save_tmp_file(image, "tess_") as (tmp_name, input_file_name):
        _run_tesseract(_input_to_cli_str(lang, config, 0, input_file_name, tmp_name))
        with open(tmp_name + ".tsv", "rb") as output_file:
            output = output_file.read().decode("utf-8")
        result: Dict[str, List[Union[str, int, float]]] = {}
        rows = [row.split("\t") for row in output.strip().split("\n")]
        if len(rows) < 2:
            return result
        header = rows.pop(0)
        length = len(header)
        if len(rows[-1]) < length:
            # Fixes bug that occurs when last text string in TSV is null, and
            # last row is missing a final cell in TSV file
            rows[-1].append("")

        str_col_idx = -1
        str_col_idx += length

        val: Union[str, int]
        for i, head in enumerate(header):
            result[head] = []
            for row in rows:
                if len(row) <= i:
                    continue

                val = row[i]
                if row[i].isdigit() and i != -1 and head != "text":
                    val = int(row[i])
                elif head == "text":
                    val = str(row[i])
                result[head].append(val)

        return result


def tesseract_line_to_detectresult(detect_result_list: List[DetectionResult]) -> List[DetectionResult]:
    """
    Generating text line DetectionResult based on Tesseract word grouping. It generates line bounding boxes from
    word bounding boxes.
    :param detect_result_list: A list of detection result
    :return: An extended list of detection result
    """

    line_detect_result: List[DetectionResult] = []
    for _, block_group_iter in groupby(detect_result_list, key=lambda x: x.block):
        block_group = []
        for _, line_group_iter in groupby(list(block_group_iter), key=lambda x: x.line):
            block_group.extend(list(line_group_iter))
        assert all(isinstance(detect_result.box, list) for detect_result in block_group)
        ulx = min(detect_result.box[0] for detect_result in block_group)  # type: ignore
        uly = min(detect_result.box[1] for detect_result in block_group)  # type: ignore
        lrx = max(detect_result.box[2] for detect_result in block_group)  # type: ignore
        lry = max(detect_result.box[3] for detect_result in block_group)  # type: ignore
        if block_group:
            line_detect_result.append(
                DetectionResult(
                    box=[ulx, uly, lrx, lry],
                    class_id=2,
                    class_name=LayoutType.line,
                    text=" ".join(
                        [detect_result.text for detect_result in block_group if isinstance(detect_result.text, str)]
                    ),
                )
            )
    if line_detect_result:
        detect_result_list.extend(line_detect_result)
    return detect_result_list


def predict_text(np_img: ImageType, supported_languages: str, text_lines: bool, config: str) -> List[DetectionResult]:
    """
    Calls tesseract directly with some given configs. Requires Tesseract to be installed.

    :param np_img: Image in np.array.
    :param supported_languages: To improve ocr extraction quality it is helpful to pre-select the language of the
                                detected text, if this in known in advance. Combinations are possible, e.g. "deu",
                                "fr+eng".
    :param text_lines: If True, it will return DetectionResults of Text lines as well.
    :param config: The config parameter passing to Tesseract. Consult also https://guides.nyu.edu/tesseract/usage
    :return: A list of tesseract extractions wrapped in DetectionResult
    """

    np_img = np_img.astype(np.uint8)
    results = image_to_dict(np_img, supported_languages, config)
    all_results = []

    for caption in zip(
        results["left"],
        results["top"],
        results["width"],
        results["height"],
        results["conf"],
        results["text"],
        results["block_num"],
        results["line_num"],
    ):
        score = float(caption[4])
        if int(score) != -1:
            word = DetectionResult(
                box=[caption[0], caption[1], caption[0] + caption[2], caption[1] + caption[3]],
                score=score / 100,
                text=caption[5],
                block=str(caption[6]),
                line=str(caption[7]),
                class_id=1,
                class_name=LayoutType.word,
            )
            all_results.append(word)
    if text_lines:
        all_results = tesseract_line_to_detectresult(all_results)
    return all_results


class TesseractOcrDetector(ObjectDetector):
    """
    Text object detector based on Tesseracts OCR engine. Note that tesseract has to be installed separately.

    The current Tesseract release is 4.1.1. A version 5.xx can be integrated via direct installation at
    https://github.com/tesseract-ocr/tesseract. Building from source is necessary here.

    Documentation can be found here: https://tesseract-ocr.github.io/

    All configuration options that are available via pytesseract can be given via the configuration. The best overview
    can be found at https://pypi.org/project/pytesseract/.

        tesseract_config_path = ModelCatalog.get_full_path_configs("dd/conf_tesseract.yaml")
        ocr_detector = TesseractOcrDetector(tesseract_config_path)

        detection_result = ocr_detector.predict(bgr_image_as_np_array)

    To use it within a pipeline

        tesseract_config_path = ModelCatalog.get_full_path_configs("dd/conf_tesseract.yaml")
        ocr_detector = TesseractOcrDetector(tesseract_config_path)

        text_extract = TextExtractionService(ocr_detector)
        pipe = DoctectionPipe([text_extract])

        df = pipe.analyze(path="path/to/dir")

        for dp in df:
            ...
    """

    def __init__(
        self,
        path_yaml: str,
        config_overwrite: Optional[List[str]] = None,
    ):
        """
        Set up the configuration which is stored in a yaml-file, that need to be passed through.

        :param path_yaml: The path to the yaml config
        :param config_overwrite: Overwrite config parameters defined by the yaml file with new values.
                                 E.g. ["oem=14"]
        """
        self.name = _TESS_PATH
        if config_overwrite is None:
            config_overwrite = []

        hyper_param_config = set_config_by_yaml(path_yaml)
        if len(config_overwrite):
            hyper_param_config.update_args(config_overwrite)

        self.path_yaml = path_yaml
        self.config_overwrite = config_overwrite
        self.config = hyper_param_config

        if self.config.LINES:
            self.categories = {"1": LayoutType.word, "2": LayoutType.line}
        else:
            self.categories = {"1": LayoutType.word}

    def predict(self, np_img: ImageType) -> List[DetectionResult]:
        """
        Transfer of a numpy array and call of pytesseract. Return of the detection results.

        :param np_img: image as numpy array
        :return: A list of DetectionResult
        """
        detection_results = predict_text(
            np_img,
            supported_languages=self.config.LANGUAGES,
            text_lines=self.config.LINES,
            config=config_to_cli_str(self.config, "LANGUAGES", "LINES"),
        )
        return detection_results

    @classmethod
    def get_requirements(cls) -> List[Requirement]:
        return [get_tesseract_requirement()]

    def clone(self) -> PredictorBase:
        return self.__class__(self.path_yaml, self.config_overwrite)

    def possible_categories(self) -> List[ObjectTypes]:
        if self.config.LINES:
            return [LayoutType.word, LayoutType.line]
        return [LayoutType.word]

    def set_language(self, language: ObjectTypes) -> None:
        """
        Pass a language to change the model selection. For runtime language selection.
        :param language: `Languages`
        """
        self.config.LANGUAGES = _LANG_CODE_TO_TESS_LANG_CODE.get(language, language.value)



# -*- coding: utf-8 -*-
# File: cats.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Categories related mapping functions. They can be set within a pipeline directly after a dataflow
builder method of a dataset.
"""

from collections import defaultdict
from typing import Dict, List, Literal, Mapping, Optional, Sequence, Tuple, Union

from ..datapoint.annotation import CategoryAnnotation, ContainerAnnotation, ImageAnnotation, SummaryAnnotation
from ..datapoint.image import Image
from ..utils.settings import ObjectTypes, TypeOrStr, get_type
from .maputils import LabelSummarizer, curry


@curry
def cat_to_sub_cat(
    dp: Image,
    categories_dict_names_as_key: Dict[TypeOrStr, str],
    cat_to_sub_cat_dict: Optional[Dict[TypeOrStr, TypeOrStr]] = None,
) -> Image:
    """
    Replace some category with its affiliated sub category of CategoryAnnotations. Suppose your category name is `foo`
    and comes along with sub_category_annotations `foo_1` and `foo_2` then this adapter will replace `foo` with
    `foo_1` or `foo_2`, respectively.

    :param dp: Image datapoint
    :param categories_dict_names_as_key: A dict of all possible categories and their ids
    :param cat_to_sub_cat_dict: e.g. {'foo': 'sub_cat_1', 'bak': 'sub_cat_2'}
    :return: Image with updated Annotations
    """

    if cat_to_sub_cat_dict is None:
        return dp
    cat_to_sub_cat_dict_obj_type = {get_type(key): get_type(value) for key, value in cat_to_sub_cat_dict.items()}
    categories_dict = categories_dict_names_as_key
    for ann in dp.get_annotation_iter(category_names=list(cat_to_sub_cat_dict_obj_type.keys())):
        sub_cat_type = cat_to_sub_cat_dict_obj_type[get_type(ann.category_name)]
        sub_cat = ann.get_sub_category(sub_cat_type)
        if sub_cat:
            ann.category_name = sub_cat.category_name
            ann.category_id = categories_dict[ann.category_name]

    return dp


@curry
def re_assign_cat_ids(dp: Image, categories_dict_name_as_key: Dict[TypeOrStr, str]) -> Image:
    """
    Re-assigning category ids is sometimes necessary to align with categories of the `DatasetCategories` . E.g.
    consider the situation where some categories are filtered. In order to guarantee alignment of category ids of the
    `DatasetCategories` the ids in the annotation have to be re-assigned.

    Annotations that as not in the dictionary provided will removed from the image.

    :param dp: Image
    :param categories_dict_name_as_key:
    :return: Image
    """

    anns_to_remove: List[ImageAnnotation] = []
    for ann in dp.get_annotation_iter():
        if ann.category_name in categories_dict_name_as_key:
            ann.category_id = categories_dict_name_as_key[ann.category_name]
        else:
            anns_to_remove.append(ann)

    for ann in anns_to_remove:
        dp.remove(ann)

    return dp


@curry
def filter_cat(
    dp: Image, categories_as_list_filtered: List[TypeOrStr], categories_as_list_unfiltered: List[TypeOrStr]
) -> Image:
    """
    Filters category annotations based on the on a list of categories to be kept and a list of all possible
    category names that might be available in dp.

    :param dp: Image datapoint
    :param categories_as_list_filtered: A list of category names with categories to keep. Using a dataset e.g.
                                        my_data.categories.get_categories(as_dict=False,filtered=True)
    :param categories_as_list_unfiltered: A list of all available category names. Using a dataset e.g.
                                          my_data.categories.get_categories(as_dict=False)
    :return: Image with filtered Annotations
    """

    cats_to_remove_list = [cat for cat in categories_as_list_unfiltered if cat not in categories_as_list_filtered]

    remove_cats_mapper = remove_cats(category_names=cats_to_remove_list)  # pylint: disable=E1120  # 259
    dp = remove_cats_mapper(dp)

    categories_dict_name_as_key = {v: str(k) for k, v in enumerate(categories_as_list_filtered, 1)}
    re_assign_cat_ids_mapper = re_assign_cat_ids(  # pylint: disable=E1120
        categories_dict_name_as_key=categories_dict_name_as_key
    )
    dp = re_assign_cat_ids_mapper(dp)

    return dp


@curry
def filter_summary(
    dp: Image, sub_cat_to_sub_cat_names_or_ids: Mapping[TypeOrStr, Sequence[TypeOrStr]], use_category_name: bool = True
) -> Optional[Image]:
    """
    Filters datapoints with given summary conditions. If several conditions are given, it will filter out datapoints
    that do not satisfy all conditions.

    :param dp: Image datapoint
    :param sub_cat_to_sub_cat_names_or_ids: A dict of list. The key correspond to the sub category key to look for in
                                            the summary. The value correspond to a sequence of either category names
                                            or category ids
    :param use_category_name: With respect to the previous argument, if set to True, it will look if the category name
                              corresponds to any of the given values. If False it will look for category ids.
    :return: Image or None
    """
    for key, values in sub_cat_to_sub_cat_names_or_ids.items():
        if use_category_name and dp.summary:
            if dp.summary.get_sub_category(get_type(key)).category_name in values:
                return dp
        elif dp.summary:
            if dp.summary.get_sub_category(get_type(key)).category_id in values:
                return dp
    return None


@curry
def image_to_cat_id(
    dp: Image,
    category_names: Optional[Union[TypeOrStr, Sequence[TypeOrStr]]] = None,
    sub_categories: Optional[Union[Mapping[TypeOrStr, TypeOrStr], Mapping[TypeOrStr, Sequence[TypeOrStr]]]] = None,
    summary_sub_category_names: Optional[Union[TypeOrStr, Sequence[TypeOrStr]]] = None,
    id_name_or_value: Literal["id", "name", "value"] = "id",
) -> Tuple[Dict[TypeOrStr, Union[List[int], List[int]]], str]:
    """
    Extracts all category_ids, sub category information or summary sub category information with given names into a
    defaultdict. This mapping is useful when running evaluation with e.g. an accuracy metric.

    **Example 1:**

        dp contains image annotations

            ImageAnnotation(category_name='foo',category_id='1',...),
            ImageAnnotation(category_name='bak',category_id='2',...),
            ImageAnnotation(category_name='baz',category_id='3',...),
            ImageAnnotation(category_name='foo',category_id='1',...),

        Then

             image_to_cat_id(category_names=['foo', 'bak', 'baz'])(dp)

        will return

            ({'foo':['1', '1'], 'bak':[ '2'], 'baz':['3']}, image_id)


    **Example 2:**

        dp contains image annotations as given in Example 1. Moreover, the 'foo' image annotation have sub categories:

            foo_sub_1: CategoryAnnotation(category_name='sub_1', category_id='4')
            foo_sub_1: CategoryAnnotation(category_name='sub_1', category_id='5')

            image_to_cat_id(sub_categories={'foo':'foo_sub_1'})

        will return

            ({'foo_sub_1':['5', '6']}, image_id)



    :param dp: Image datapoint
    :param category_names: A list of category names
    :param sub_categories: A dict {'cat':'sub_cat'} or a list. Will dump the results with sub_cat as key
    :param id_name_or_value: Only relevant for sub categories. It will extract the sub category id, the name or, if the
                             sub category is a container, it will extract a value.
    :param summary_sub_category_names: A list of summary sub categories
    :return: A defaultdict of lists
    """

    cat_container = defaultdict(list)

    if isinstance(category_names, str):
        category_names = [category_names]
    if not category_names:
        category_names = []

    if isinstance(summary_sub_category_names, str):
        summary_sub_category_names = [summary_sub_category_names]
    if not summary_sub_category_names:
        summary_sub_category_names = []

    tmp_sub_category_names: Dict[str, Sequence[str]] = {}

    if sub_categories is not None:
        for key, val in sub_categories.items():
            if isinstance(val, str):
                val = [val]
            tmp_sub_category_names[key] = val

    if id_name_or_value not in ("id", "name", "value"):
        raise ValueError(f"id_name_or_value must be in ('id', 'name', 'value') but is {id_name_or_value}")

    if category_names or sub_categories:  # pylint: disable=R1702
        for ann in dp.get_annotation_iter():
            if ann.category_name in category_names:
                cat_container[ann.category_name].append(int(ann.category_id))
            if ann.category_name in tmp_sub_category_names:
                for sub_cat_name in tmp_sub_category_names[ann.category_name]:
                    sub_cat = ann.get_sub_category(get_type(sub_cat_name))
                    if sub_cat is not None:
                        if id_name_or_value == "id":
                            cat_container[sub_cat_name].append(int(sub_cat.category_id))
                        if id_name_or_value == "name":
                            cat_container[sub_cat_name].append(sub_cat.category_name)  # type: ignore
                        if id_name_or_value == "value":
                            if not isinstance(sub_cat, ContainerAnnotation):
                                raise ValueError(
                                    f"sub category {sub_cat_name} does not have a ContainerAnnotation. Choose another"
                                    f"value for argument id_name_or_value"
                                )
                            cat_container[sub_cat_name].append(sub_cat.value)  # type: ignore

    if dp.summary is not None and summary_sub_category_names:
        for sub_cat_name in summary_sub_category_names:
            sub_cat = dp.summary.get_sub_category(get_type(sub_cat_name))
            if id_name_or_value == "id":
                cat_container[sub_cat_name].append(int(sub_cat.category_id))
            if id_name_or_value == "name":
                cat_container[sub_cat_name].append(sub_cat.category_name)  # type: ignore
            if id_name_or_value == "value":
                if not isinstance(sub_cat, ContainerAnnotation):
                    raise ValueError(
                        f"sub category {sub_cat_name} does not have a ContainerAnnotation. Choose another"
                        f"value for argument id_name_or_value"
                    )
                cat_container[sub_cat_name].append(sub_cat.value)  # type: ignore

    return cat_container, dp.image_id


@curry
def remove_cats(
    dp: Image,
    category_names: Optional[Union[TypeOrStr, Sequence[TypeOrStr]]] = None,
    sub_categories: Optional[Union[Mapping[TypeOrStr, TypeOrStr], Mapping[TypeOrStr, Sequence[TypeOrStr]]]] = None,
    relationships: Optional[Union[Mapping[TypeOrStr, TypeOrStr], Mapping[TypeOrStr, Sequence[TypeOrStr]]]] = None,
    summary_sub_categories: Optional[Union[TypeOrStr, Sequence[TypeOrStr]]] = None,
) -> Image:
    """
    Remove categories according to given category names or sub category names. Note that these will change the container
    in which the objects are stored.

    :param dp: A datapoint image
    :param category_names: A single category name or a list of categories to remove. On default will remove
                           nothing.
    :param sub_categories: A dict with category names and a list of their sub categories to be removed
    :param relationships: A dict with category names and a list of relationship names to be removed
    :param summary_sub_categories: A single sub category or a list of sub categories from a summary to be removed
    :return: A datapoint image with removed categories
    """

    if isinstance(category_names, str):
        category_names = [category_names]
    if category_names is None:
        category_names = []

    if sub_categories is None:
        sub_categories = {}

    if relationships is None:
        relationships = {}

    if isinstance(summary_sub_categories, str):
        summary_sub_categories = [summary_sub_categories]

    anns_to_remove = []

    for ann in dp.get_annotation_iter():
        if ann.category_name in category_names:
            anns_to_remove.append(ann)
        if ann.category_name in sub_categories.keys():
            sub_cats_to_remove = sub_categories[ann.category_name]
            if isinstance(sub_cats_to_remove, str):
                sub_cats_to_remove = [sub_cats_to_remove]
            for sub_cat in sub_cats_to_remove:
                ann.remove_sub_category(get_type(sub_cat))
        if ann.category_name in relationships.keys():
            relationships_to_remove = relationships[ann.category_name]
            if isinstance(relationships_to_remove, str):
                relationships_to_remove = [relationships_to_remove]
            for relation in relationships_to_remove:
                ann.remove_relationship(key=get_type(relation))

    for ann in anns_to_remove:
        dp.remove(ann)

    if summary_sub_categories is not None:
        if dp.summary is not None:
            for sub_cat in summary_sub_categories:
                dp.summary.remove_sub_category(get_type(sub_cat))

    return dp


@curry
def add_summary(dp: Image, categories: Mapping[str, ObjectTypes]) -> Image:
    """
    Adding a summary with the number of categories in an image.

    :param dp: Image
    :param categories: A dict of all categories, e.g. `{"1": "text", "2":"title", ...}`
    :return: Image
    """
    category_list = list(categories.values())
    anns = dp.get_annotation(category_names=category_list)
    summarizer = LabelSummarizer(categories)
    for ann in anns:
        summarizer.dump(ann.category_id)
    summary_dict = summarizer.get_summary()
    summary = SummaryAnnotation()
    for cat_id, val in summary_dict.items():
        summary.dump_sub_category(
            categories[cat_id], CategoryAnnotation(category_name=categories[cat_id], category_id=str(val))
        )
    dp.summary = summary
    return dp



# -*- coding: utf-8 -*-
# File: __init__.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Simple framework inspired by

<https://huggingface.co/docs/datasets/>

for creating datasets.

Create an info card, a DataFlowBaseBuilder derived instance, possibly a category card and a
DatasetBase derived instance to create a data set.
"""

from ..utils.file_utils import pytorch_available
from .base import *
from .dataflow_builder import DataFlowBaseBuilder
from .info import *
from .instances import *
from .registry import *
from .save import *

if pytorch_available():
    from .adapter import *



# -*- coding: utf-8 -*-
# File: publaynet.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Module for Publaynet dataset. Place the dataset as follows

    publaynet
    ├── test
    │ ├── PMC_1.png
    ├── train
    │ ├── PMC_2.png
    ├── val
    │ ├── PMC_3.png
    ├── train.json
    ├── val.json
"""


from typing import Mapping, Union

from ...dataflow import DataFlow, MapData, MapDataComponent
from ...dataflow.custom_serialize import SerializerCoco
from ...mapper.cats import add_summary, filter_cat
from ...mapper.cocostruct import coco_to_image
from ...utils.settings import DatasetType, LayoutType
from ..base import _BuiltInDataset
from ..dataflow_builder import DataFlowBaseBuilder
from ..info import DatasetCategories, DatasetInfo
from ..registry import dataset_registry

_NAME = "publaynet"
_DESCRIPTION = (
    "PubLayNet is a dataset for document layout analysis. It contains images of research papers and "
    "articles \n"
    " and annotations for various elements in a page such as “text”, “list”, “figure” etc in these "
    "research \n"
    " paper images. The dataset was obtained by automatically matching the XML representations and the \n"
    " content of over 1 million PDF articles that are publicly available on PubMed Central. \n"
)
_LICENSE = (
    "The annotations in this dataset belong to IBM and are licensed under a Community Data License Agreement \n"
    " – Permissive – Version 1.0 License. IBM does not own the copyright of the images. Use of the images \n"
    " must abide by the PMC Open Access Subset Terms of Use."
)
_URL = (
    "https://dax-cdn.cdn.appdomain.cloud/dax-publaynet/1.0.0/"
    "publaynet.tar.gz?_ga=2.23017467.1796315263.1628754613-1173244232.1625045842"
)
_SPLITS: Mapping[str, str] = {"train": "train", "val": "val"}
_TYPE = DatasetType.object_detection

_LOCATION = "publaynet"

_ANNOTATION_FILES: Mapping[str, str] = {"train": "train.json", "val": "val.json"}
_INIT_CATEGORIES = [LayoutType.text, LayoutType.title, LayoutType.list, LayoutType.table, LayoutType.figure]


@dataset_registry.register("publaynet")
class Publaynet(_BuiltInDataset):
    """
    Publaynet
    """

    _name = _NAME

    @classmethod
    def _info(cls) -> DatasetInfo:
        return DatasetInfo(name=_NAME, description=_DESCRIPTION, license=_LICENSE, url=_URL, splits=_SPLITS, type=_TYPE)

    def _categories(self) -> DatasetCategories:
        return DatasetCategories(init_categories=_INIT_CATEGORIES)

    def _builder(self) -> "PublaynetBuilder":
        return PublaynetBuilder(location=_LOCATION, annotation_files=_ANNOTATION_FILES)


class PublaynetBuilder(DataFlowBaseBuilder):
    """
    Publaynet dataflow builder
    """

    def build(self, **kwargs: Union[str, int]) -> DataFlow:
        """
        Returns a dataflow from which you can stream datapoints of images. The following arguments affect the returns
        of the dataflow:

        `split:` Split of the dataset. Can be `train`,`val` or `test`. Default: `val`

        `max_datapoints:` Will stop iterating after max_datapoints. Default: `None`

        `load_image:` Will load the image for each datapoint.  Default: `False`

        `fake_score:` Will add a fake score so that annotations look like predictions

        :return: dataflow
        """
        split = str(kwargs.get("split", "val"))
        max_datapoints = kwargs.get("max_datapoints")
        if max_datapoints is not None:
            max_datapoints = int(max_datapoints)
        load_image = kwargs.get("load_image", False)
        fake_score = kwargs.get("fake_score", False)

        # Load
        path = self.get_workdir() / self.get_annotation_file(split)
        df = SerializerCoco.load(path, max_datapoints=max_datapoints)

        # Map
        df = MapDataComponent(df, lambda dp: self.get_workdir() / self.get_split(split) / dp, "file_name")
        coco_mapper = coco_to_image(  # pylint: disable=E1120  # 259
            self.categories.get_categories(init=True),
            load_image,
            filter_empty_image=True,
            fake_score=fake_score,
        )
        df = MapData(df, coco_mapper)

        if self.categories.is_filtered():
            df = MapData(
                df,
                filter_cat(  # pylint: disable=E1120
                    self.categories.get_categories(as_dict=False, filtered=True),
                    self.categories.get_categories(as_dict=False, filtered=False),
                ),
            )

        df = MapData(df, add_summary(self.categories.get_categories(filtered=True)))
        return df



# -*- coding: utf-8 -*-
# File: test_prodigystruct.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing the module mapper.prodigystruct
"""

from math import isclose
from typing import Dict

from pytest import mark

from deepdoctection.datapoint.image import Image
from deepdoctection.mapper import image_to_prodigy, prodigy_to_image
from deepdoctection.utils.detection_types import JsonDict

from .data import DatapointImage, DatapointProdigy


@mark.basic
def test_prodigy_to_image(
    datapoint_prodigy: JsonDict, categories_prodigy: Dict[str, str], prodigy_results: DatapointProdigy
) -> None:
    """
    testing prodigy_to_image is mapping correctly
    """

    load_image = True
    # Act
    prodigy_to_image_mapper = prodigy_to_image(categories_prodigy, load_image, False)  # pylint: disable=E1120  # 259
    dp = prodigy_to_image_mapper(datapoint_prodigy)

    # Assert
    datapoint = prodigy_results
    assert dp is not None
    test_anns = dp.get_annotation()
    assert len(test_anns) == datapoint.get_number_anns()
    assert dp.width == datapoint.get_width(load_image)
    assert dp.height == datapoint.get_height(load_image)
    assert test_anns[0].category_name == datapoint.get_first_ann_category(False)
    assert test_anns[0].category_id == datapoint.get_first_ann_category(True)
    assert test_anns[0].bounding_box
    assert isclose(test_anns[0].bounding_box.ulx, datapoint.get_first_ann_box().ulx, rel_tol=1e-15)
    assert isclose(test_anns[0].bounding_box.uly, datapoint.get_first_ann_box().uly, rel_tol=1e-15)
    assert isclose(test_anns[0].bounding_box.width, datapoint.get_first_ann_box().w, rel_tol=1e-15)
    assert isclose(test_anns[0].bounding_box.height, datapoint.get_first_ann_box().h, rel_tol=1e-15)


@mark.basic
def test_image_to_prodigy(datapoint_image: Image, image_results: DatapointImage) -> None:
    """
    testing image_to_prodigy is mapping correctly
    """

    # Act
    output = image_to_prodigy(datapoint_image)

    # Assert
    datapoint = image_results
    assert output["image"] == datapoint.get_image_str()
    assert output["text"] == datapoint.get_text()
    assert len(output["spans"]) == datapoint.get_len_spans()
    assert output["spans"][0] == datapoint.get_first_span()
    assert output["spans"][1] == datapoint.get_second_span()



# -*- coding: utf-8 -*-
# File: test_tp_frcnn_train.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module train.tp_frcnn_train
"""
from typing import Any
from unittest.mock import MagicMock, patch

from pytest import mark

from deepdoctection.datasets import DatasetBase
from deepdoctection.utils.file_utils import tensorpack_available, tf_available
from deepdoctection.utils.metacfg import set_config_by_yaml

from ..test_utils import collect_datapoint_from_dataflow, set_num_gpu_to_one

if tf_available() and tensorpack_available():
    from deepdoctection.extern.tp.tpfrcnn.config.config import model_frcnn_config
    from deepdoctection.train.tp_frcnn_train import get_train_dataflow, train_faster_rcnn


@mark.requires_tf
@patch("deepdoctection.mapper.tpstruct.os.path.isfile", MagicMock(return_value=True))
@patch("deepdoctection.train.tp_frcnn_train.set_mp_spawn")
def test_get_train_dataflow(
    set_mp_spawn: Any, test_dataset: DatasetBase, path_to_tp_frcnn_yaml: str  # pylint: disable=W0613
) -> None:
    """
    test get train dataflow
    """

    # Arrange
    config = set_config_by_yaml(path_to_tp_frcnn_yaml)
    categories = test_dataset.dataflow.categories.get_categories(filtered=True)
    model_frcnn_config(config, categories)

    # Act
    df = get_train_dataflow(test_dataset, config, False)
    df_list = collect_datapoint_from_dataflow(df)

    # Assert
    assert len(df_list) == 2
    dp = df_list[0]
    assert "image" in dp
    assert dp["image"].ndim == 3
    assert "gt_boxes" in dp
    assert "gt_labels" in dp


@mark.requires_tf
@patch("deepdoctection.mapper.tpstruct.os.path.isfile", MagicMock(return_value=True))
@patch("deepdoctection.extern.tp.tpcompat.get_num_gpu", MagicMock(side_effect=set_num_gpu_to_one))
@patch("deepdoctection.extern.tp.tpfrcnn.config.config.get_num_gpu", MagicMock(side_effect=set_num_gpu_to_one))
@patch("deepdoctection.train.tp_frcnn_train.ModelSaver")
@patch("deepdoctection.train.tp_frcnn_train.PeriodicCallback")
@patch("tensorpack.utils.logger.set_logger_dir")
@patch("deepdoctection.train.tp_frcnn_train.launch_train_with_config")
@patch("deepdoctection.train.tp_frcnn_train.set_mp_spawn")
def test_train_faster_rcnn(
    set_mp_spawn: Any,  # pylint: disable=W0613
    patch_launch_train_with_config: Any,
    patch_set_logger_dir: Any,  # pylint: disable=W0613
    patch_periodic_cb: Any,  # pylint: disable=W0613
    patch_model_saver: Any,  # pylint: disable=W0613
    path_to_tp_frcnn_yaml: str,
    test_dataset: DatasetBase,
) -> None:
    """
    test train faster rcnn runs until "launch_train_with_config"
    """

    # Arrange
    train_faster_rcnn(
        path_config_yaml=path_to_tp_frcnn_yaml,
        dataset_train=test_dataset,
        path_weights="",
        log_dir="/test/log_dir",
        dataset_val=test_dataset,
        metric_name="coco",
        pipeline_component_name="ImageLayoutService",
    )

    # Assert
    patch_launch_train_with_config.assert_called_once()



# -*- coding: utf-8 -*-
# File: base.py

# Copyright (c) Tensorpack Contributors
# Licensed under the Apache License, Version 2.0 (the "License")

"""
This file replaces relevant parts of the Dataflow package. Most of the code has been taken from

<https://github.com/tensorpack/dataflow/blob/master/dataflow/dataflow/base.py>
"""

import threading
from abc import ABC, abstractmethod
from typing import Any, Iterator, no_type_check

from ..utils.utils import get_rng


class DataFlowTerminated(BaseException):
    """
    An exception indicating that the DataFlow is unable to produce any more
    data, i.e. something wrong happened so that calling `__iter__`
    cannot give a valid iterator anymore.
    In most DataFlow this will never be raised.
    """


class DataFlowResetStateNotCalled(BaseException):
    """
    An exception indicating that `reset_state()` has not been called before starting
    iteration.
    """

    def __init__(self) -> None:
        super().__init__("Iterating a dataflow requires .reset_state() to be called first")


class DataFlowReentrantGuard:
    """
    A tool to enforce non-reentrancy.
    Mostly used on DataFlow whose `get_data` is stateful,
    so that multiple instances of the iterator cannot co-exist.
    """

    def __init__(self) -> None:
        self._lock = threading.Lock()

    def __enter__(self) -> None:
        self._succ = self._lock.acquire(False)
        if not self._succ:
            raise threading.ThreadError("This dataflow is not reentrant!")

    @no_type_check
    def __exit__(self, exc_type, exc_val, exc_tb):
        self._lock.release()
        return False


class DataFlow:
    """Base class for all DataFlow"""

    @abstractmethod
    def __iter__(self) -> Iterator[Any]:
        """
        * A dataflow is an iterable. The `__iter__` method should yield a list or dict each time.
          Note that dict is **partially** supported at the moment: certain dataflow does not support dict.
        * The `__iter__` method can be either finite (will stop iteration) or infinite
          (will not stop iteration). For a finite dataflow, `__iter__` can be called
          again immediately after the previous call returned.
        * For many dataflow, the `__iter__` method is non-reentrant, which means for an dataflow
          instance ``df``, `df.__iter__` cannot be called before the previous
          `df.__iter__` call has finished (iteration has stopped).
          When a dataflow is non-reentrant, `df.__iter__` should throw an exception if
          called before the previous call has finished.
          For such non-reentrant dataflows, if you need to use the same dataflow in two places,
          you need to create two dataflow instances.
        Yields:
            list/dict: The datapoint, i.e. list/dict of components.
        """

    def __len__(self) -> int:
        """
        * A dataflow can optionally implement `__len__`. If not implemented, it will
          throw `NotImplementedError`.
        * It returns an integer representing the size of the dataflow.
          The return value **may not be accurate or meaningful** at all.
          When saying the length is "accurate", it means that
          `__iter__` will always yield this many of datapoints before it stops iteration.
        * There could be many reasons why `__len__` is inaccurate.
          For example, some dataflow has dynamic size, if it throws away datapoints on the fly.
          Some dataflow mixes the datapoints between consecutive passes over
          the dataset, due to parallelism and buffering.
          In this case it does not make sense to stop the iteration anywhere.
        * Due to the above reasons, the length is only a rough guidance.
          And it's up to the user how to interpret it.
          Inside tensorpack it's only used in these places:
          + A default ``steps_per_epoch`` in training, but you probably want to customize
            it yourself, especially when using data-parallel trainer.
          + The length of progress bar when processing a dataflow.
          + Used by `InferenceRunner` to get the number of iterations in inference.
            In this case users are **responsible** for making sure that `__len__` is "accurate".
            This is to guarantee that inference is run on a fixed set of images.
        Returns:
            int: rough size of this dataflow.
        Raises:
            `NotImplementedError` if this DataFlow doesn't have a size.
        """
        raise NotImplementedError

    def reset_state(self) -> None:
        """
        * The caller must guarantee that `reset_state` should be called **once and only once**
          by the **process that uses the dataflow** before `__iter__` is called.
          The caller thread of this method should stay alive to keep this dataflow alive.
        * It is meant for certain initialization that involves processes,
          e.g., initialize random number generators (RNG), create worker processes.
          Because it's very common to use RNG in data processing,
          developers of dataflow can also subclass `RNGDataFlow` to have easier access to
          a properly-initialized RNG.
        * A dataflow is not fork-safe after `reset_state` is called (because this will violate the guarantee).
          There are a few other dataflows that are not fork-safe anytime, which will be mentioned in the docs.
        * You should take the responsibility and follow the above guarantee if you're the caller of a dataflow yourself
          (either when you're using dataflow outside tensorpack, or if you're writing a wrapper dataflow).
        * Tensorpack's built-in forking dataflows (`MultiProcessRunner`, `MultiProcessMapData`, etc)
          and other component that uses dataflows (`InputSource`)
          already take care of the responsibility of calling this method.
        """
        raise NotImplementedError


class RNGDataFlow(DataFlow, ABC):
    """A DataFlow with RNG"""

    rng = None
    """
    `self.rng` is a `np.random.RandomState` instance that is initialized
    correctly (with different seeds in each process) in `RNGDataFlow.reset_state()`.
    """

    def reset_state(self) -> None:
        """Reset the RNG"""
        self.rng = get_rng(self)


class ProxyDataFlow(DataFlow):
    """Base class for DataFlow that proxies another.
    Every method is proxied to ``self.df`` unless overriden by a subclass.
    """

    def __init__(self, df: DataFlow) -> None:
        """
        :param df: DataFlow to proxy.
        """
        self.df = df

    def reset_state(self) -> None:
        self.df.reset_state()

    def __len__(self) -> int:
        return self.df.__len__()

    def __iter__(self) -> Iterator[Any]:
        return self.df.__iter__()



# -*- coding: utf-8 -*-
# File: test_convert.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module datapoint.convert
"""

from pytest import mark

from deepdoctection.datapoint import convert_pdf_bytes_to_np_array_v2

from .conftest import TestPdfPage


@mark.basic
def test_convert_pdf_bytes_to_np_array_v2(pdf_page: TestPdfPage) -> None:
    """
    testing convert_pdf_bytes_to_np_array_v2 returns a np.array of correct shape
    """

    # Act
    np_array = convert_pdf_bytes_to_np_array_v2(pdf_page.pdf_bytes)

    # Assert
    assert np_array.shape == pdf_page.np_array_shape_default


@mark.basic
def test_convert_pdf_bytes_to_np_array_v2_with_dpi_300(pdf_page: TestPdfPage) -> None:
    """
    testing convert_pdf_bytes_to_np_array_v2 returns a np.array of correct shape when dpi is set to 300
    """

    # Act
    np_array = convert_pdf_bytes_to_np_array_v2(pdf_page.pdf_bytes, 300)

    # Assert
    assert np_array.shape == pdf_page.np_array_shape



# -*- coding: utf-8 -*-
# File: test_layouttest.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module datasets.instances.layouttest
"""

from pytest import mark

from deepdoctection.datasets import LayoutTest

from ...test_utils import collect_datapoint_from_dataflow, get_test_path


@mark.basic
def test_dataset_layouttest_returns_image() -> None:
    """
    test dataset layouttest returns image
    """

    # Arrange
    layouttest = LayoutTest()
    layouttest.dataflow.get_workdir = get_test_path  # type: ignore
    layouttest.dataflow.splits = {"test": ""}
    layouttest.dataflow.annotation_files = {"test": "test_layout.jsonl"}
    df = layouttest.dataflow.build()

    # Act & Assert
    df_list = collect_datapoint_from_dataflow(df)
    assert len(df_list) == 2


@mark.basic
def test_dataset_layouttest_with_load_image_returns_image() -> None:
    """
    test dataset layouttest returns image
    """

    # Arrange
    layouttest = LayoutTest()
    layouttest.dataflow.get_workdir = get_test_path  # type: ignore
    layouttest.dataflow.splits = {"test": ""}
    layouttest.dataflow.annotation_files = {"test": "test_layout.jsonl"}
    df = layouttest.dataflow.build(load_image=True)

    # Act & Assert
    df_list = collect_datapoint_from_dataflow(df)
    assert len(df_list) == 2
    assert df_list[0].image is not None



# -*- coding: utf-8 -*-
# File: test_text.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Testing module pipe.text
"""

from typing import List
from unittest.mock import MagicMock

from pytest import mark, raises

from deepdoctection.datapoint import BoundingBox, Image, ImageAnnotation
from deepdoctection.extern.base import DetectionResult, ObjectDetector, PdfMiner
from deepdoctection.pipe.text import TextExtractionService, TextOrderService
from deepdoctection.utils.settings import CellType, LayoutType, Relationships


class TestTextExtractionService:
    """
    Test TextExtractionService2. In this setting, extraction will be tested when extract_from_category is None.
    """

    def setup_method(self) -> None:
        """
        setup necessary components
        """

        self._text_extract_detector = MagicMock(spec=ObjectDetector)
        self._text_extract_detector.name = "mock_text_extractor"
        self.text_extraction_service = TextExtractionService(self._text_extract_detector)

    @mark.basic
    def test_integration_pipeline_component(
        self, dp_image_fully_segmented_fully_tiled: Image, word_detect_result: List[DetectionResult]
    ) -> None:
        """
        Integration test through calling `pass_datapoint` of pipeline component
        """

        # Arrange
        self._text_extract_detector.predict = MagicMock(return_value=word_detect_result)

        # Act
        dp = self.text_extraction_service.pass_datapoint(dp_image_fully_segmented_fully_tiled)
        anns = dp.get_annotation(category_names=LayoutType.word)

        # Assert
        first_text_ann = anns[0]

        embedding_bbox = first_text_ann.image.get_embedding(dp.image_id)  # type: ignore
        assert embedding_bbox == first_text_ann.bounding_box


class TestTextExtractionServiceWithPdfPlumberDetector:
    """
    Test TextExtractionServiceWithPdfPlumberDetector with PdfPlumberTextDetector
    """

    def setup_method(self) -> None:
        """
        setup necessary components
        """

        self._text_extract_detector = MagicMock(spec=PdfMiner)
        self._text_extract_detector.name = "mock_pdfminer"
        self.text_extraction_service = TextExtractionService(self._text_extract_detector)

    @mark.basic
    def test_integration_pipeline_component(
        self, dp_image_fully_segmented_fully_tiled: Image, word_detect_result: List[DetectionResult]
    ) -> None:
        """
        Integration test through calling `pass_datapoint` of pipeline component
        """

        # Arrange
        dp_image_fully_segmented_fully_tiled.pdf_bytes = b"some bytes"
        self._text_extract_detector.predict = MagicMock(return_value=word_detect_result)
        self._text_extract_detector.get_width_height = MagicMock(return_value=(600, 400))

        # Act
        dp = self.text_extraction_service.pass_datapoint(dp_image_fully_segmented_fully_tiled)
        anns = dp.get_annotation(category_names=LayoutType.word)

        # Assert
        first_text_ann = anns[0]

        embedding_bbox = first_text_ann.image.get_embedding(dp.image_id)  # type: ignore
        assert embedding_bbox == first_text_ann.bounding_box


@mark.basic
def test_text_extraction_service_raises_error_with_inconsistent_attributes() -> None:
    """
    Testing TextExtractionService does not build when instantiating with a PdfMiner and passing some ROI
    """

    # Arrange
    text_extract_detector = MagicMock(spec=PdfMiner)
    text_extract_detector.name = "mock_pdfminer"

    # Act and Assert
    with raises(TypeError):
        TextExtractionService(text_extract_detector, extract_from_roi=LayoutType.table)


class TestTextExtractionServiceWithSubImage:
    """
    Test TextExtractionService where extract_from_category = "table". As the unit test will not differ from test
    class above only an integration test will be executed.
    """

    def setup_method(self) -> None:
        """
        setup necessary components
        """

        self._text_extract_detector = MagicMock(spec=ObjectDetector, accepts_batch=False)
        self._text_extract_detector.name = "mock_text_extractor"
        self.text_extraction_service = TextExtractionService(
            self._text_extract_detector, extract_from_roi=LayoutType.table
        )

    @mark.basic
    def test_integration_pipeline_component(
        self,
        dp_image_with_layout_anns: Image,
        double_word_detect_results: List[List[DetectionResult]],
        word_box_global: List[BoundingBox],
    ) -> None:
        """
        integration test through calling  serve of pipeline component
        """

        # Arrange
        self._text_extract_detector.predict = MagicMock(side_effect=double_word_detect_results)

        # Act
        dp = self.text_extraction_service.pass_datapoint(dp_image_with_layout_anns)
        word_anns = dp.get_annotation(category_names=LayoutType.word)
        table_anns = dp.get_annotation(category_names=LayoutType.table)

        assert len(word_anns) == 4
        assert len(table_anns) == 2

        assert isinstance(table_anns, list) and isinstance(word_anns, list)
        first_table_ann = table_anns[0]
        second_table_ann = table_anns[1]
        first_word_ann = word_anns[0]
        second_word_ann = word_anns[1]
        third_word_ann = word_anns[2]
        fourth_word_ann = word_anns[3]

        global_box_fta = first_word_ann.image.get_embedding(dp.image_id)  # type: ignore
        assert global_box_fta == word_box_global[0]
        local_box_fta = first_word_ann.image.get_embedding(first_table_ann.annotation_id)  # type: ignore
        assert local_box_fta == first_word_ann.bounding_box
        ft_text_ann = first_table_ann.image.get_annotation(annotation_ids=first_word_ann.annotation_id)[  # type: ignore
            0
        ]

        global_box_sta = second_word_ann.image.get_embedding(dp.image_id)  # type: ignore
        assert global_box_sta == word_box_global[1]
        local_box_sta = second_word_ann.image.get_embedding(first_table_ann.annotation_id)  # type: ignore
        assert local_box_sta == second_word_ann.bounding_box
        ft_text_ann = first_table_ann.image.get_annotation(  # type: ignore
            annotation_ids=second_word_ann.annotation_id
        )[0]
        assert isinstance(ft_text_ann, ImageAnnotation)

        global_box_tta = third_word_ann.image.get_embedding(dp.image_id)  # type: ignore
        assert global_box_tta == word_box_global[2]
        local_box_tta = third_word_ann.image.get_embedding(second_table_ann.annotation_id)  # type: ignore
        assert local_box_tta == third_word_ann.bounding_box
        st_text_ann = second_table_ann.image.get_annotation(  # type: ignore
            annotation_ids=third_word_ann.annotation_id
        )[0]
        assert isinstance(st_text_ann, ImageAnnotation)

        global_box_fta = fourth_word_ann.image.get_embedding(dp.image_id)  # type: ignore
        assert global_box_fta == word_box_global[3]
        local_box_fta = fourth_word_ann.image.get_embedding(second_table_ann.annotation_id)  # type: ignore
        assert local_box_fta == fourth_word_ann.bounding_box
        st_text_ann = second_table_ann.image.get_annotation(  # type: ignore
            annotation_ids=fourth_word_ann.annotation_id
        )[0]
        assert isinstance(st_text_ann, ImageAnnotation)


class TestTextOrderService:
    """
    Test TextOrderService
    """

    @staticmethod
    @mark.basic
    def test_integration_pipeline_component(dp_image_with_layout_and_word_annotations: Image) -> None:
        """
        test integration_pipeline_component
        """

        # Arrange
        text_order_service = TextOrderService(
            text_container=LayoutType.word,
            floating_text_block_names=[LayoutType.title, LayoutType.text, LayoutType.list],
            text_block_names=[
                LayoutType.title,
                LayoutType.text,
                LayoutType.list,
                LayoutType.cell,
                CellType.header,
                CellType.body,
            ],
        )
        dp_image = dp_image_with_layout_and_word_annotations

        # Act
        text_order_service.pass_datapoint(dp_image)

        # Assert
        layout_anns = dp_image.get_annotation(category_names=[LayoutType.title, LayoutType.text])
        word_anns = dp_image.get_annotation(category_names=LayoutType.word)

        # only need to check on layout_anns and word_anns, if sub cats have been added
        # and numbers are correctly assigned

        sub_cat = layout_anns[0].get_sub_category(Relationships.reading_order)
        assert sub_cat.category_id == "1"
        sub_cat = layout_anns[1].get_sub_category(Relationships.reading_order)
        assert sub_cat.category_id == "2"

        sub_cat = word_anns[0].get_sub_category(Relationships.reading_order)
        assert sub_cat.category_id == "1"
        sub_cat = word_anns[1].get_sub_category(Relationships.reading_order)
        assert sub_cat.category_id == "2"
        sub_cat = word_anns[2].get_sub_category(Relationships.reading_order)
        assert sub_cat.category_id == "1"
        sub_cat = word_anns[3].get_sub_category(Relationships.reading_order)
        assert sub_cat.category_id == "2"



# -*- coding: utf-8 -*-
# File: conftest.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.


"""
Fixtures for mapper package testing
"""
from typing import Any, Dict, Mapping, Optional, Union

from pytest import fixture

from deepdoctection.datapoint import Image
from deepdoctection.utils.detection_types import ImageType, JsonDict
from deepdoctection.utils.settings import LayoutType, ObjectTypes

from .data import (
    DatapointCoco,
    DatapointImage,
    DatapointPageDict,
    DatapointProdigy,
    DatapointPubtabnet,
    DatapointXfund,
    IIITar13KJson,
)


@fixture(name="datapoint_coco")
def fixture_datapoint_coco() -> Dict[str, Any]:
    """
    Datapoint as received from SerializerCoco
    """

    return DatapointCoco().dp


@fixture(name="categories_coco")
def fixture_categories_coco() -> Mapping[str, ObjectTypes]:
    """
    Categories as Dict
    """
    return DatapointCoco().categories


def get_coco_white_image(path: str, type_id: str = "np") -> Optional[Union[str, ImageType]]:
    """
    Returns a white image
    :param path: An image path
    :param type_id: "np" or "b64"
    :return:
    """
    return DatapointCoco().get_white_image(path, type_id)


@fixture(name="coco_results")
def fixture_coco_results() -> DatapointCoco:
    """
    DatapointCoco
    """
    return DatapointCoco()


@fixture(name="datapoint_pubtabnet")
def fixture_datapoint_pubtabnet() -> Dict[str, Any]:
    """
    Datapoint as received from SerializerCoco
    """

    return DatapointPubtabnet().dp


@fixture(name="categories_name_as_key_pubtabnet")
def fixture_categories_name_as_key_pubtabnet() -> Mapping[ObjectTypes, str]:
    """
    Categories as Dict
    """
    return DatapointPubtabnet().categories_as_names


@fixture(name="pubtabnet_results")
def fixture_pubtabnet_results() -> DatapointPubtabnet:
    """
    DatapointPubtabnet
    """
    return DatapointPubtabnet()


def get_pubtabnet_white_image(path: str, type_id: str = "np") -> Optional[Union[str, ImageType]]:
    """
    Returns a white image
    :param path: An image path
    :param type_id: "np" or "b64"
    """

    if path == DatapointPubtabnet().dp["filename"]:
        return DatapointPubtabnet().get_white_image(path, type_id)
    return None


def get_always_pubtabnet_white_image(path: str, type_id: str = "np") -> Optional[Union[str, ImageType]]:
    """
    Returns a white image
    :param path: An image path
    :param type_id: "np" or "b64"
    """

    return DatapointPubtabnet().get_white_image(path, type_id)


def get_always_pubtabnet_white_image_from_bytes(
    pdf_bytes: str, dpi: Optional[int] = None
) -> Optional[Union[str, ImageType]]:
    """
    Returns a white image
    """
    if pdf_bytes and dpi is not None:
        pass
    return DatapointPubtabnet().get_white_image("", "np")


def get_always_bytes(path: str) -> bytes:
    """
    Returns bytes
    """
    if path:
        pass
    return b"\x04\x00"


@fixture(name="datapoint_prodigy")
def fixture_datapoint_prodigy() -> JsonDict:
    """
    Datapoint as received from Prodigy db
    """

    return DatapointProdigy().dp


@fixture(name="categories_prodigy")
def fixture_categories_prodigy() -> Mapping[ObjectTypes, str]:
    """
    Categories as Dict
    """
    return DatapointProdigy().categories  # type: ignore


def get_datapoint_prodigy() -> DatapointProdigy:
    """
    DatapointProdigy
    """
    return DatapointProdigy()


@fixture(name="prodigy_results")
def fixture_prodigy_results() -> DatapointProdigy:
    """
    DatapointProdigy
    """
    return DatapointProdigy()


@fixture(name="datapoint_image")
def fixture_datapoint_image() -> Image:
    """
    Image
    """
    return DatapointImage().image


@fixture(name="datapoint_image_with_summary")
def fixture_datapoint_image_with_summary() -> Image:
    """
    Image with summary annotation
    """
    return DatapointImage().get_image_with_summary()


@fixture(name="page_dict")
def fixture_page_dict() -> JsonDict:
    """
    page file
    """
    return DatapointPageDict().get_page_dict()


@fixture(name="datapoint_xfund")
def fixture_datapoint_xfund() -> Dict[str, Any]:
    """
    Datapoint as received from Xfund dataset
    """

    return DatapointXfund().dp  # type: ignore


@fixture(name="xfund_category_names")
def fixture_xfund_category_names() -> Mapping[str, ObjectTypes]:
    """
    Xfund category names mapping
    """

    return DatapointXfund().get_category_names_mapping()


@fixture(name="xfund_category_dict")
def fixture_xfund_category_dict_name_as_key() -> Mapping[LayoutType, str]:
    """
    Xfund category dict name as key
    """
    return DatapointXfund().get_categories_dict()


@fixture(name="layoutlm_input")
def fixture_layoutlm_input() -> JsonDict:
    """
    Layoutlm input
    """
    return DatapointXfund().get_layout_input()


@fixture(name="layoutlm_v2_input")
def fixture_layoutlm_v2_input() -> JsonDict:
    """
    Layoutlm_v2 input
    """
    return DatapointXfund().get_layout_v2_input()


@fixture(name="raw_layoutlm_features")
def fixture_raw_layoutlm_featurest() -> JsonDict:
    """
    Layoutlm input
    """
    return DatapointXfund().get_raw_layoutlm_features()


@fixture(name="xfund_categories_dict_name_as_key")
def fixture_xfund_categories_dict_name_as_key() -> Mapping[ObjectTypes, str]:
    """
    categories_dict_name_as_key
    """
    return DatapointXfund().get_categories_dict_names_as_key()


@fixture(name="ner_token_to_id_mapping")
def fixture_ner_token_to_id_mapping() -> Mapping[ObjectTypes, Any]:
    """
    ner_token_to_id_mapping
    """
    return DatapointXfund().get_net_token_to_id_mapping()


@fixture(name="datapoint_iiitar13kjson")
def fixture_datapoint_iiitar13kjson() -> Dict[str, Any]:
    """
    Datapoint as received from iiitar13k dataset already converted into json format
    """

    return IIITar13KJson().dp


@fixture(name="iiitar13k_categories_name_as_keys")
def fixture_iiitar13k_categories_name_as_keys() -> Mapping[ObjectTypes, str]:
    """
    iiitar13k category names dict
    """

    return IIITar13KJson().get_categories_name_as_keys()


@fixture(name="iiitar13k_category_names_mapping")
def fixture_xfund_category_names_mapping() -> Mapping[str, ObjectTypes]:
    """
    iiitar13k category names mapping
    """

    return IIITar13KJson().get_category_names_mapping()


@fixture(name="iiitar13k_results")
def fixture_iiitar13k_results() -> IIITar13KJson:
    """
    iiitar13k results
    """
    return IIITar13KJson()



# -*- coding: utf-8 -*-
# File: annotation.py

# Copyright 2021 Dr. Janis Meyer. All rights reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""
Dataclass for annotations and their derived classes.
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Any, Dict, List, Optional, Union, no_type_check

from ..utils.detection_types import JsonDict
from ..utils.identifier import get_uuid, is_uuid_like
from ..utils.logger import logger
from ..utils.settings import DefaultType, ObjectTypes, SummaryType, TypeOrStr, get_type
from .box import BoundingBox
from .convert import as_dict


@no_type_check
def ann_from_dict(cls, **kwargs):
    """
    A factory function to create subclasses of annotations from a given dict
    """
    ann = cls(kwargs.get("external_id"), kwargs.get("category_name"), kwargs.get("category_id"), kwargs.get("score"))
    ann.active = kwargs.get("active")
    ann._annotation_id = kwargs.get("_annotation_id")  # pylint: disable=W0212
    if isinstance(kwargs.get("sub_categories"), dict):
        for key, value in kwargs["sub_categories"].items():
            if "value" in value:
                ann.dump_sub_category(key, ContainerAnnotation.from_dict(**value))
            else:
                ann.dump_sub_category(key, CategoryAnnotation.from_dict(**value))
    if isinstance(kwargs.get("relationships"), dict):
        for key, values in kwargs["relationships"].items():
            for value in values:
                ann.dump_relationship(key, value)
    return ann


@dataclass  # type: ignore
class Annotation(ABC):
    """
    Abstract base class for all types of annotations. This abstract base class only implements general methods for
    correctly assigning annotation_ids. It also has an active flag which is set to True. Only active annotations will be
    returned when querying annotations from image containers.

    Annotation id should never be assigned by yourself. One possibility of assigning an id depending on an external
    value is to set an external id, which in turn creates an annotation id as a function of this.
    An annotation id can only be explicitly set, provided the value is a md5 hash.

    Note that otherwise ids will be generated automatically if the annotation object is dumped in a parent container,
    either an image or annotation (e.g. sub-category). If no id is supplied, the annotation id is created depending
    on the defining attributes (key and value pairs) as specified in the return value of
    `get_defining_attributes`.

    `active`: Always set to `True`. You can change the value using `deactivate` .

    `external_id`: A string or integer value for generating an annotation id. Note, that the resulting annotation
    id will not depend on the defining attributes.

    `_annotation_id`: Unique id for annotations. Will always be given as string representation of a md5-hash.
    """

    active: bool = field(default=True, init=False, repr=True)
    external_id: Optional[Union[str, int]] = field(default=None, init=True, repr=False)
    _annotation_id: Optional[str] = field(default=None, init=False, repr=True)

    def __post_init__(self) -> None:
        """
        Will check, if the external id provided is an uuid. If not will use the external id as seed for defining an
        uuid.
        """

        if self.external_id is not None:
            external_id = str(self.external_id)
            if is_uuid_like(external_id):
                self.annotation_id = external_id
            else:
                self.annotation_id = get_uuid(external_id)
        self._assert_attributes_have_str()

    @property
    def annotation_id(self) -> str:
        """
        annotation_id
        """
        if self._annotation_id:
            return self._annotation_id
        raise ValueError("Dump annotation first or pass external_id to create an annotation id")

    @annotation_id.setter
    def annotation_id(self, input_id: str) -> None:
        """
        annotation_id setter
        """
        if self._annotation_id is not None:
            raise AssertionError("Annotation_id already defined and cannot be reset")
        if is_uuid_like(input_id):
            self._annotation_id = input_id
        elif isinstance(input_id, property):
            pass
        else:
            raise ValueError("Annotation_id must be uuid3 string")

    @abstractmethod
    def get_defining_attributes(self) -> List[str]:
        """
        Defining attributes of an annotation instance are attributes, of which you think that they uniquely
        describe the annotation object. If you do not provide an external id, only the defining attributes will be used
        for generating the annotation id.

        :return: A list of attributes.
        """
        raise NotImplementedError

    def _assert_attributes_have_str(self, state_id: bool = False) -> None:
        defining_attributes = self.get_state_attributes() if state_id else self.get_defining_attributes()
        for attr in defining_attributes:
            if not hasattr(eval("self." + attr), "__str__"):  # pylint: disable=W0123
                raise AttributeError(f"Attribute {attr} must have __str__ method")

    @staticmethod
    def set_annotation_id(annotation: "CategoryAnnotation", *container_id_context: Optional[str]) -> str:
        """
        Defines the `annotation_id` by attributes of the annotation class as well as by external parameters given by a
        tuple or list of container id contexts.

        :param annotation: The annotation instance for which the id should be generated.
        :param container_id_context: A tuple/list of strings on which you want the resulting annotation id to depend on.
        :return: A uuid that uniquely characterizes the annotation.
        """
        for container_id in container_id_context:
            if container_id is not None:
                assert is_uuid_like(container_id), f"{container_id} is not a uuid"
        attributes = annotation.get_defining_attributes()
        attributes_values = [str(getattr(annotation, attribute)) for attribute in attributes]
        return get_uuid(*attributes_values, *container_id_context)  # type: ignore

    def as_dict(self) -> Dict[str, Any]:
        """
        Returning the full dataclass as dict. Uses the custom `convert.as_dict` to disregard attributes defined by
        `remove_keys`.

        :return: A custom dict.
        """

        img_dict = as_dict(self, dict_factory=dict)

        return img_dict

    def deactivate(self) -> None:
        """
        Sets `active` to False. When calling `Image.get_annotations()` it will be filtered.
        """
        self.active = False

    @classmethod
    @abstractmethod
    def from_dict(cls, **kwargs: JsonDict) -> "Annotation":
        """
        Method to initialize a derived class from dict.

        :param kwargs: dict with `Annotation` attributes

        :return: Annotation instance
        """
        raise NotImplementedError

    @staticmethod
    @abstractmethod
    def get_state_attributes() -> List[str]:
        """
        Similar to `get_defining_attributes` but for `state_id`

        :return: A list of attributes.
        """
        raise NotImplementedError

    @property
    def state_id(self) -> str:
        """
        Different to `annotation_id` this id does depend on every defined state attributes and might therefore change
        over time.

        :return: Annotation state instance
        """
        container_ids = []
        attributes = self.get_state_attributes()
        for attribute in attributes:
            attr = getattr(self, attribute)
            if isinstance(attr, dict):
                for key, value in attr.items():
                    if isinstance(value, Annotation):
                        container_ids.extend([key, value.state_id])
                    elif isinstance(value, list):
                        container_ids.extend([str(element) for element in value])
                    else:
                        raise TypeError(f"Cannot determine __str__ or annotation_id for element in {attribute}")
            elif isinstance(attr, list):
                for element in attr:
                    if isinstance(element, Annotation):
                        container_ids.append(element.state_id)
                    if isinstance(element, str):
                        container_ids.append(element)
                    else:
                        container_ids.append(str(element))
            elif hasattr(attr, "state_id"):
                container_ids.append(attr.state_id)
            else:
                container_ids.append(str(attr))
        return get_uuid(self.annotation_id, *container_ids)


@dataclass
class CategoryAnnotation(Annotation):
    """
    A general class for storing categories (labels/classes) as well as sub categories (sub-labels/subclasses),
    relationships and prediction scores.

    Sub-categories and relationships are stored in a dict, which are populated via the `dum_sub_category` or
    `dump_relationship`. If a key is already available as a sub-category, it must be explicitly removed using the
    `remove_sub_category` before replacing the sub-category.

    Note that subcategories are only accepted as category annotations. Relationships, on the other hand, are only
    managed by passing the annotation id.

    `category_name`: String will be used for selecting specific annotations. Use upper case strings.

    `category_id`: When setting a value will accept strings and ints. Will be stored as string.

    `score`: Score of a prediction.

    `sub_categories`: Do not access the dict directly. Rather use the access `get_sub_category` resp.
    `dump_sub_category`.

    `relationships`: Do not access the dict directly either. Use `get_relationship` or
    `dump_relationship` instead.
    """

    category_name: TypeOrStr = field(default=DefaultType.default_type)
    _category_name: ObjectTypes = field(default=DefaultType.default_type, init=False)
    category_id: str = field(default="")
    score: Optional[float] = field(default=None)
    sub_categories: Dict[ObjectTypes, "CategoryAnnotation"] = field(default_factory=dict, init=False, repr=True)
    relationships: Dict[ObjectTypes, List[str]] = field(default_factory=dict, init=False, repr=True)

    @property  # type: ignore
    def category_name(self) -> ObjectTypes:
        """category name"""
        return self._category_name

    @category_name.setter
    def category_name(self, category_name: TypeOrStr) -> None:
        """category name setter"""
        if not isinstance(category_name, property):
            self._category_name = get_type(category_name)

    def __post_init__(self) -> None:
        self.category_id = str(self.category_id)
        assert self.category_name
        self._assert_attributes_have_str(state_id=True)
        super().__post_init__()

    def dump_sub_category(
        self, sub_category_name: TypeOrStr, annotation: "CategoryAnnotation", *container_id_context: Optional[str]
    ) -> None:
        """
        Storage of sub-categories. Since sub-categories usually only depend on very few attributes and the parent
        category cannot yet be stored in a comprehensive container, it is possible to include a context of the
        annotation id in order to ensure that the sub-category annotation id is unambiguously created.

        :param sub_category_name: key for defining the sub category.
        :param annotation: Annotation instance to dump
        :param container_id_context: Tuple/list of context ids.
        """

        if sub_category_name in self.sub_categories:
            raise KeyError(f"{sub_category_name} as sub category already defined for " f"{self.annotation_id}")

        if self._annotation_id is not None:
            if annotation._annotation_id is None:  # pylint: disable=W0212
                annotation.annotation_id = self.set_annotation_id(annotation, self.annotation_id, *container_id_context)
        else:
            tmp_annotation_id = self.set_annotation_id(self)
            if annotation._annotation_id is None:  # pylint: disable=W0212
                annotation.annotation_id = annotation.set_annotation_id(
                    annotation, tmp_annotation_id, *container_id_context
                )
        self.sub_categories[get_type(sub_category_name)] = annotation

    def get_sub_category(self, sub_category_name: ObjectTypes) -> "CategoryAnnotation":
        """
        Return a sub category by its key.

        :param sub_category_name: The key of the sub-category.

        :return: sub category as CategoryAnnotation
        """
        return self.sub_categories[sub_category_name]

    def remove_sub_category(self, key: ObjectTypes) -> None:
        """
        Removes a sub category with a given key. Necessary to call, when you want to replace an already dumped sub
        category.

        :param key: A key to a sub category.
        """

        if key in self.sub_categories:
            self.sub_categories.pop(key)

    def dump_relationship(self, key: TypeOrStr, annotation_id: str) -> None:
        """
        Dumps an `annotation_id` to a given key, in order to store relations between annotations. Note, that the
        referenced annotation must be stored elsewhere.

        :param key: The key, where to place the annotation id.
        :param annotation_id: An annotation id
        """
        if not is_uuid_like(annotation_id):
            raise ValueError("Annotation_id must be uuid")

        key_type = get_type(key)
        if key not in self.relationships:
            self.relationships[key_type] = []
        if annotation_id not in self.relationships[key_type]:
            self.relationships[key_type].append(annotation_id)

    def get_relationship(self, key: ObjectTypes) -> List[str]:
        """
        Returns a list of annotation ids stored with a given relationship key.

        :param key: The key for the required relationship.
        :return: Get a (possibly) empty list of annotation ids.
        """
        if key in self.relationships:
            return self.relationships[key]
        return []

    def remove_relationship(self, key: ObjectTypes, annotation_ids: Optional[Union[List[str], str]] = None) -> None:
        """
        Remove relationship by some given keys and ids. If no annotation ids are provided all relationship according
        to the key will be removed.

        :param key: A relationship key.
        :param annotation_ids: A single annotation_id or a list. Will remove only the relation with given
                               annotation_ids, provided not None is passed as argument.
        """

        if annotation_ids is not None:
            if isinstance(annotation_ids, str):
                annotation_ids = [annotation_ids]
            for ann_id in annotation_ids:
                try:
                    self.relationships[key].remove(ann_id)
                except ValueError:
                    logger.warning("Relationship %s cannot be removed because it does not exist", key)
        else:
            self.relationships[key].clear()

    def get_defining_attributes(self) -> List[str]:
        return ["category_name", "category_id"]

    @staticmethod
    def remove_keys() -> List[str]:
        """
        A list of attributes to suspend from as_dict creation.

        :return: List of attributes.
        """
        return []

    @classmethod
    def from_dict(cls, **kwargs: JsonDict) -> "CategoryAnnotation":
        category_ann = ann_from_dict(cls, **kwargs)
        return category_ann

    @staticmethod
    def get_state_attributes() -> List[str]:
        return ["active", "sub_categories", "relationships"]


@dataclass
class ImageAnnotation(CategoryAnnotation):
    """
    A general class for storing annotations related to object detection tasks. In addition to the inherited attributes,
    the class contains a bounding box and an image attribute. The image attribute is optional and is suitable for
    generating an image from the annotation and then saving it there. Compare with the method `image.Image.
    image_ann_to_image`, which naturally populates this attribute.

    `bounding_box`: Regarding the coordinate system, if you have to define a prediction, use the system of the
    image where the object has been detected.

    `image`: Image, defined by the bounding box and cropped from its parent image. Populate this attribute with
    `Image.image_ann_to_image`.
    """

    bounding_box: Optional[BoundingBox] = field(default=None)
    image: Optional["Image"] = field(default=None, init=False, repr=False)  # type: ignore

    def get_defining_attributes(self) -> List[str]:
        return ["category_name", "bounding_box"]

    @classmethod
    def from_dict(cls, **kwargs: JsonDict) -> "ImageAnnotation":
        image_ann = ann_from_dict(cls, **kwargs)
        if box_kwargs := kwargs.get("bounding_box"):
            image_ann.bounding_box = BoundingBox.from_dict(**box_kwargs)
        return image_ann

    @staticmethod
    def get_state_attributes() -> List[str]:
        return ["active", "sub_categories", "relationships", "image"]


@dataclass
class SummaryAnnotation(CategoryAnnotation):
    """
    A dataclass for adding summaries. The various summaries can be stored as sub categories.

    Summary annotations should be stored in the attribute provided: `image.Image.summary`  and should not be
    dumped as a category.
    """

    def __post_init__(self) -> None:
        self._category_name = SummaryType.summary
        super().__post_init__()

    @classmethod
    def from_dict(cls, **kwargs: JsonDict) -> "SummaryAnnotation":
        summary_ann = ann_from_dict(cls, **kwargs)
        summary_ann.category_name = SummaryType.summary
        return summary_ann


@dataclass
class ContainerAnnotation(CategoryAnnotation):
    """
    A dataclass for transporting values along with categorical attributes. Use these types of annotations as special
    types of sub categories.

     value: Attribute to store the value. Use strings.
    """

    value: Optional[Union[List[str], str]] = field(default=None)

    def get_defining_attributes(self) -> List[str]:
        return ["category_name", "value"]

    @classmethod
    def from_dict(cls, **kwargs: JsonDict) -> "SummaryAnnotation":
        container_ann = ann_from_dict(cls, **kwargs)
        container_ann.value = kwargs.get("value")
        return container_ann
